"""Работа с компонентами Windows"""
import json
import os
import subprocess
import win32com.client
import winreg
from .core import ms
from .path import PATH_TYPES, path2str
from typing import *
shell: dict[str, win32com.client.CDispatch] = {}
shell["WScript"] = win32com.client.Dispatch("WScript.Shell")
_names = {}
_names["lnk"] = {"args": "Arguments", "cwd": "WorkingDirectory", "desc": "Description", "hotkey": "Hotkey", "icon": "IconLocation", "lnk": "FullName", "target": "TargetPath"}


def read_lnk(path: PATH_TYPES) -> dict:
  """Прочитать ярлык .lnk"""
  r = {}
  lnk = shell["WScript"].CreateShortcut(path2str(path, True).replace("/", "\\"))
  r["src"] = lnk
  for k, v in _names["lnk"].items():
    r[k] = getattr(lnk, v, None)
  return r


def write_lnk(path: PATH_TYPES, target: str, args: str = None, cwd: str = None, desc: str = None, hotkey: str = None, icon: str = None):
  """Создать ярлык .lnk"""
  lnk = shell["WScript"].CreateShortCut(path2str(path, True).replace("/", "\\"))
  lnk.TargetPath = target.replace("/", "\\")
  if args != None:
    lnk.Arguments = args
  if cwd != None:
    lnk.WorkingDirectory = cwd.replace("/", "\\")
  if desc != None:
    lnk.Description = desc
  if hotkey != None:
    lnk.Hotkey = hotkey
  if icon != None:
    lnk.IconLocation = icon.replace("/", "\\")
  lnk.Save()
  return lnk


def hide_file(path: PATH_TYPES, *, recursive: bool = False, unhide: bool = False):
  """Скрыть файл используя системную команду `attrib`"""
  args = ["attrib"]
  args.append("-H" if unhide else "+H")
  args.append(path2str(path, True).replace("/", "\\"))
  if recursive:
    args += ["/S", "/D"]
  subprocess.run(args, check=True)


def unhide_file(path: PATH_TYPES, *, recursive: bool = False, hide: bool = False):
  """Противоположность функции скрытия файла"""
  return hide_file(path, recursive=recursive, unhide=not hide)


def path2win(path: PATH_TYPES, **kw) -> str:
  kw.setdefault("to_abs", True)
  return path2str(path, **kw).replace("/", "\\")


class LnkFile:
  """Открыть/создать файл `.lnk`"""

  def __init__(self, path: PATH_TYPES, readonly: bool = True):
    path = path2win(path)
    if readonly:
      if not os.path.isfile(path):
        raise ms.file.NotAFileError(path)
    self.obj: win32com.client.CDispatch = shell["WScript"].CreateShortcut(path)
    self.readonly = readonly

  def __getitem__(self, k):
    result = getattr(self.obj, k, None)
    if result == "":
      return None
    return result

  def __setitem__(self, k, v):
    if self.readonly:
      raise RuntimeError("Read-only shortcut")
    if v is None:
      v = ""
    setattr(self.obj, k, v)

  def save(self):
    """Сохранить изменения"""
    if self.readonly:
      raise RuntimeError("Read-only shortcut")
    self.obj.Save()

  @property
  def args_str(self) -> None | str:
    return self["Arguments"]

  @args_str.setter
  def args_str(self, v):
    self["Arguments"] = v

  @property
  def args(self) -> None | list[str]:
    line = self.args_str
    if line is None:
      return
    import shlex
    return shlex.split(line, posix=False)

  @args.setter
  def args(self, v):
    if v is None:
      self["Arguments"] = None
      return
    import shlex
    self["Arguments"] = shlex.join(v)

  @property
  def description(self) -> None | str:
    return self["Description"]

  @description.setter
  def description(self, v):
    self["Description"] = v

  @property
  def hot_key(self) -> None | str:
    return self["Hotkey"]

  @hot_key.setter
  def hot_key(self, v):
    self["Hotkey"] = v

  @property
  def icon_loc(self) -> None | str:
    return self["IconLocation"]

  @icon_loc.setter
  def icon_loc(self, v):
    self["IconLocation"] = v

  @property
  def path(self) -> str:
    return self["FullName"]

  @property
  def target_path(self) -> None | str:
    return self["TargetPath"]

  @target_path.setter
  def target_path(self, v):
    self["TargetPath"] = path2win(v)

  @property
  def window_style(self) -> int:
    return self["WindowStyle"]

  @window_style.setter
  def window_style(self, v):
    self["WindowStyle"] = v

  @property
  def working_dir(self) -> None | str:
    return self["WorkingDirectory"]

  @working_dir.setter
  def working_dir(self, v):
    self["WorkingDirectory"] = path2win(v)

# XXX


# Названия и номера локаций
REG_LOC_BY_NAME = {
    "~": winreg.HKEY_CURRENT_USER,
    "CLASSES_ROOT": winreg.HKEY_CLASSES_ROOT,
    "CURRENT_CONFIG": winreg.HKEY_CURRENT_CONFIG,
    "CURRENT_USER": winreg.HKEY_CURRENT_USER,
    "DYN_DATA": winreg.HKEY_DYN_DATA,
    "HKCC": winreg.HKEY_CURRENT_CONFIG,
    "HKCR": winreg.HKEY_CLASSES_ROOT,
    "HKCU": winreg.HKEY_CURRENT_USER,
    "HKDD": winreg.HKEY_DYN_DATA,
    "HKLM": winreg.HKEY_LOCAL_MACHINE,
    "HKPD": winreg.HKEY_PERFORMANCE_DATA,
    "HKU": winreg.HKEY_USERS,
    "LOCAL_MACHINE": winreg.HKEY_LOCAL_MACHINE,
    "PERFORMANCE_DATA": winreg.HKEY_PERFORMANCE_DATA,
    "USERS": winreg.HKEY_USERS,
}
REG_LOC_BY_NUM = {}
# Названия окружений
ENV_SYSTEM = "system"
ENV_SYSTEM64 = "system64"
ENV_USER = "user"
# Добавить локации с исходными названиями
for k in dir(winreg):
  if k.startswith("HKEY_"):
    v = getattr(winreg, k)
    REG_LOC_BY_NAME[k] = v
    REG_LOC_BY_NUM[v] = k
# Часто используемые локации


class RegFastLocations:
  SYSTEM_WINDOWS = "/HKLM/SOFTWARE/Microsoft/Windows/CurrentVersion"
  SYSTEM64_WINDOWS = "/HKLM/SOFTWARE/WOW6432Node/Microsoft/Windows/CurrentVersion"
  USER_WINDOWS = "/HKCU/SOFTWARE/Microsoft/Windows/CurrentVersion"
  #
  SYSTEM_AUTORUN_ONCE = SYSTEM_WINDOWS + "/RunOnce"
  SYSTEM_AUTORUN = SYSTEM_WINDOWS + "/Run"
  SYSTEM_ENV = "HKLM/SYSTEM/CurrentControlSet/Control/Session Manager/Environment"
  SYSTEM_PROG_LIST = SYSTEM_WINDOWS + "/Uninstall"
  SYSTEM64_AUTORUN_ONCE = SYSTEM64_WINDOWS + "/RunOnce"
  SYSTEM64_AUTORUN = SYSTEM64_WINDOWS + "/Run"
  SYSTEM64_PROG_LIST = SYSTEM64_WINDOWS + "/Uninstall"
  USER_AUTORUN_ONCE = USER_WINDOWS + "/RunOnce"
  USER_AUTORUN = USER_WINDOWS + "/Run"
  USER_ENV = "HKCU/Environment"
  USER_PROG_LIST = USER_WINDOWS + "/Uninstall"
# Функции для работы с путями


def reg_prep_path(path: str, lstrip=False) -> str:
  if not path:
    return "" if lstrip else "\\"
  path = path.replace("/", "\\").rstrip("\\")
  if not path:
    return "" if lstrip else "\\"
  if lstrip:
    return path.lstrip("\\")
  return path


def reg_path2loc(path: str) -> tuple[int, str]:
  path = reg_prep_path(path)
  if path.startswith("~"):
    path = "\\" + path
  assert path.startswith("\\")
  parts = path.split("\\", 2)
  return REG_LOC_BY_NAME[parts[1].upper()], ("\\" if len(parts) == 2 else parts[2])


def reg_loc2path(loc: int | str, loc_path: str = None) -> str:
  if isinstance(loc, int):
    loc = REG_LOC_BY_NUM[loc]
  return ("\\" + loc.upper() + "\\" + reg_prep_path(loc_path, True)).rstrip("\\")
# Различные варианты для работы с значениями реестра


class RegFolder:
  """Папка реестра. С обязательным указанием типов значений"""

  def __init__(self, loc: int | str, loc_path: str, **kw):
    if isinstance(loc, str):
      loc = REG_LOC_BY_NAME[loc.upper()]
    assert loc in REG_LOC_BY_NUM
    self._loc_num = loc
    self._loc_path = loc_path
    self._opened_key = None
    self._path = None
    self._use_count = 0
    self.open_kw = kw

  def __enter__(self):
    if not self._opened_key:
      self._opened_key = winreg.OpenKeyEx(self.loc_num, self.loc_path + "\\", **self.open_kw)
    self._use_count += 1
    return self

  def __exit__(self, *a):
    self._use_count -= 1
    if self._opened_key:  # Если ключ открыт
      if self._use_count < 1:  # и никем не используется
        self._opened_key.Close()  # он будет закрыт

  def __contains__(self, k) -> bool:
    with self:
      try:
        winreg.QueryValueEx(self._opened_key, k)
      except FileNotFoundError:
        return False
    return True

  def __delitem__(self, k):
    with self:
      winreg.DeleteValue(self._opened_key, k)

  def __getitem__(self, k):
    """`value,type`"""
    with self:
      try:
        return winreg.QueryValueEx(self._opened_key, k)
      except FileNotFoundError:
        raise KeyError(k)

  def __setitem__(self, k, v):
    if not isinstance(v, (list, tuple)):
      raise ValueError("Value must be a tuple with a value and a type")
    if not len(v) == 2:
      raise ValueError("Value must be a tuple with a value and a type")
    if not isinstance(v[1], int):
      raise ValueError("Second value must be the registry value type")
    value, type = v
    with self:
      winreg.SetValueEx(self._opened_key, k, 0, type, value)

  @property
  def auto(self):
    """Тип значений определяется автоматически"""
    return RegFolderAuto(self.loc_num, self.loc_path, **self.open_kw)

  @property
  def json(self):
    """Все значения в реестре хранятся в виде JSON"""
    return RegFolderJson(self.loc_num, self.loc_path, **self.open_kw)

  @property
  def loc_name(self) -> str:
    """Название локации в реестре (пример: `HKEY_CURRENT_USER`)"""
    return REG_LOC_BY_NUM[self._loc_num]

  @property
  def loc_num(self) -> int:
    """Номер локации в реестре"""
    return self._loc_num

  @property
  def loc_path(self) -> str:
    """Путь без названия локации (пример: `SOFTWARE\\Microsoft\\Windows`)"""
    return self._loc_path

  @property
  def loc_tuple(self) -> tuple[int, str]:
    """Номер локации и путь раздельно (`loc_num` и `loc_path`)"""
    return self._loc_num, self._loc_path

  @property
  def normal(self):
    return self

  @property
  def path(self) -> str:
    """Путь с названием локации (пример: `\\HKEY_CURRENT_USER\\SOFTWARE\\Microsoft\\Windows`)"""
    if self._path is None:
      self._path = reg_loc2path(self._loc_num, self._loc_path)
    return self._path

  @classmethod
  def create(cls, loc, loc_path, **kw):
    self = cls(loc, loc_path, **kw)
    self._opened_key = winreg.CreateKeyEx(self.loc_num, self.loc_path + "\\", **kw)
    return self

  @classmethod
  def create_from_path(cls, path: str, **kw):
    return cls.create(*reg_path2loc(path), **kw)

  @classmethod
  def from_path(cls, path: str, **kw):
    return cls(*reg_path2loc(path), **kw)

  def get(self, key: str, default=None):
    with self:
      try:
        return winreg.QueryValueEx(self._opened_key, k)
      except FileNotFoundError:
        return default

  def setdefault(self, key, value, type: int = None):
    if not key in self:
      self[key] = value if type is None else (value, type)

  def pop(self, key: str, default=Ellipsis):
    try:
      value = winreg.QueryValueEx(self._opened_key, k)
      del self[key]
    except FileNotFoundError:
      if default is Ellipsis:
        raise KeyError(key)
      value = default
    return value


class RegFolderAuto(RegFolder):
  """Папка реестра. Тип значений определяется автоматически"""

  def __getitem__(self, k):
    with self:
      try:
        return winreg.QueryValueEx(self._opened_key, k)[0]
      except FileNotFoundError:
        raise KeyError(k)

  def __setitem__(self, k, v):
    if isinstance(v, float):
      int_v = int(v)
      if v == int_v:
        v = int_v
    with self:
      winreg.SetValueEx(self._opened_key, k, 0, self._detect_type(v), v)

  def _detect_type(self, v):
    if v is None:
      return winreg.REG_NONE
    if isinstance(v, bytes):
      return winreg.REG_BINARY
    if isinstance(v, float):
      raise ValueError("float cannot be written to the registry")
    if isinstance(v, int):
      if v < 0:
        raise ValueError("Number must be positive")
      if v > 0xffffffff:
        if v > 0xffffffffffffffff:
          raise ValueError("Number is too big")
        return winreg.REG_QWORD
      return winreg.REG_DWORD
    if isinstance(v, str):
      return winreg.REG_MULTI_SZ if "\n" in v else winreg.REG_SZ
    raise ValueError("Failed to detect type")

  @property
  def auto(self):
    return self

  @property
  def normal(self):
    return RegFolder(self.loc_num, self.loc_path, **self.open_kw)

  def setdefault(self, key: str, value):
    if not key in self:
      self[key] = value


class RegFolderJson(RegFolderAuto):
  """Папка реестра. Все значения в реестре хранятся в виде JSON"""

  def __getitem__(self, k):
    return json.loads(RegFolderAuto.__getitem__(self, k))

  def __setitem__(self, k, v, **kw):
    kw.setdefault("separators", (",", ":"))
    RegFolderAuto.__setitem__(self, k, json.dumps(v, **kw))

  @property
  def json(self):
    return self
# Проводник реестра (с поддержкой рабочей папки)


class RegExplorer:
  def __init__(self, cwd: str = None):
    self.cwd = cwd
    self.folder_cls: type[RegFolder | RegFolderAuto | RegFolderJson] = RegFolder
    self._opened_keys = {}

  def _resolve(self, path) -> tuple[int, str]:
    path = reg_prep_path(path) if path else ""
    if not path.startswith("\\"):
      path = self.cwd + "\\" + path
    return reg_path2loc(path)

  @property
  def cwd(self) -> str:
    """Current Working Dir"""
    return self._cwd

  @cwd.setter
  def cwd(self, v):
    if v is None:
      self._cwd = "\\"
    else:
      v = reg_prep_path(v) if v else ""
      assert v.startswith("\\")
      self._cwd = v

  def chdir(self, path):
    self.cwd = self._resolve(path)

  def create_folder(self, path, **kw):
    if isinstance(path, RegFolder):
      return path
    return self.folder_cls.create(*self._resolve(path), **kw)

  def open_folder(self, path, **kw):
    if isinstance(path, RegFolder):
      return path
    return self.folder_cls(*self._resolve(path), **kw)

  def get_value(self, path, key, **kw):
    return self.open_folder(path, **kw)[key]

  def set_value(self, path, key, value, type=None, *, mkdir=False, **kw):
    if not type is None:
      value = value, type
    (self.create_folder(path, **kw) if mkdir else self.open_folder(path, **kw))[key] = value

  def del_value(self, path, key, **kw):
    del self.open_folder(path, **kw)[key]
# Утилиты


class utils:
  @staticmethod
  def add_autorun(env: str, name: str, cmd: str, *, expand=True, once=False):
    """Зарегистрировать программу для автозапуска"""
    path = getattr(RegFastLocations, env.upper() + "AUTORUN")
    if once:
      path += "Once"
    RegFolder.create_from_path(path)[name] = cmd, (winreg.REG_EXPAND_SZ if expand else winreg.REG_SZ)

  @staticmethod
  def add_system_autorun(name: str, cmd: str, **kw):
    utils.add_autorun(ENV_SYSTEM, name, cmd, **kw)

  @staticmethod
  def add_system64_autorun(name: str, cmd: str, **kw):
    utils.add_autorun(ENV_SYSTEM64, name, cmd, **kw)

  @staticmethod
  def add_user_autorun(name: str, cmd: str, **kw):
    utils.add_autorun(ENV_USER, name, cmd, **kw)



import configparser
from io import StringIO
from typing import Any
from .core import ms


@ms.utils.generator2list(tuple)
def parse_list(line: str) -> list[str]:
  for i in line.split(";"):
    i = i.strip()
    if i:
      yield i


class DesktopFile:
  """Создать файл .desktop"""

  def __init__(self, path: str):
    self.path = ms.path.Path(path)
    self.parser = configparser.ConfigParser(interpolation=None)

  def __getitem__(self, key: str | tuple[str, str] | tuple[str, str, Any]):
    if isinstance(key, str):
      return self.parser[key]
    if len(key) == 2:
      return self.parser.get(key[0], key[1])
    if len(key) == 3:
      return self.parser.get(key[0], key[1], fallback=key[2])
    raise TypeError("Key must be str or tuple[str,str] or tuple[str,str,Any]")

  def __setitem__(self, key: str | tuple[str, str], value: Any):
    if isinstance(key, str):
      self.parser[key] = value
    elif len(key) == 2:
      if not self.parser.has_section(key[0]):
        self.parser.add_section(key[0])
      self.parser.set(key[0], key[1], value)
    else:
      raise TypeError("Key must be str or tuple[str,str]")

  def read(self):
    self.parser.read(self.path.path, "utf-8")

  @classmethod
  def load(cls, path: str):
    self = cls(path)
    self.read()
    return self

  def save(self):
    tmp_path = self.path.path + ms.file.TMP_SUFFIX
    with ms.path.TempFiles(tmp_path) as tmp:
      with open(tmp_path, "w", encoding="utf-8") as f:
        self.parser.write(f)
      ms.file.delete(self.path)
      ms.file.move(tmp_path, self.path)

  def to_string(self):
    with StringIO() as f:
      self.parser.write(f)
      f.seek(0)
      return f.read()

  def save2(self):
    return ms.file.write(self.path, self.to_string())

  @property
  def name(self) -> None | str:
    return self["Desktop Entry", "Name"]

  @name.setter
  def name(self, v):
    self["Desktop Entry", "Name"] = v

  @property
  def exec(self) -> None | str:
    return self["Desktop Entry", "Exec"]

  @exec.setter
  def exec(self, v):
    self["Desktop Entry", "Exec"] = v

  @property
  def icon(self) -> None | str:
    return self["Desktop Entry", "Icon", None]

  @icon.setter
  def icon(self, v):
    self["Desktop Entry", "Icon"] = v

  @property
  def terminal(self) -> bool:
    value = self["Desktop Entry", "Terminal", False]
    if isinstance(value, bool):
      return value
    return value.lower() == "true"

  @terminal.setter
  def terminal(self, v):
    self["Desktop Entry", "Terminal"] = "true" if v else "false"

  @property
  def keywords(self):
    return parse_list(self["Desktop Entry", "Keywords", ""])

  @keywords.setter
  def keywords(self, v):
    if isinstance(v, str):
      self["Desktop Entry", "Keywords"] = v
    else:
      self["Desktop Entry", "Keywords"] = ";".join(v)

  @property
  def actions(self):
    return parse_list(self["Desktop Entry", "Actions", ""])

  @actions.setter
  def actions(self, v):
    if isinstance(v, str):
      self["Desktop Entry", "Actions"] = v
    else:
      self["Desktop Entry", "Actions"] = ";".join(v)

  @property
  def categories(self):
    return parse_list(self["Desktop Entry", "Categories", ""])

  @categories.setter
  def categories(self, v):
    if isinstance(v, str):
      self["Desktop Entry", "Categories"] = v
    else:
      self["Desktop Entry", "Categories"] = ";".join(v)

  @property
  def mimetypes(self):
    return parse_list(self["Desktop Entry", "MimeType", ""])

  @mimetypes.setter
  def mimetypes(self, v):
    if isinstance(v, str):
      self["Desktop Entry", "MimeType"] = v
    else:
      self["Desktop Entry", "MimeType"] = ";".join(v)



import os
import sys
from . import _module_info
from datetime import datetime
from functools import cached_property
from logging import Logger
from time import time
try:
  from datetime import timezone
except Exception:
  timezone = None


def _get_main_file():
  def func():
    if getattr(sys, "frozen", False):
      return sys.executable
    if "__main__" in sys.modules:
      if hasattr(sys.modules["__main__"], "__file__"):
        return sys.modules["__main__"].__file__
  result = func()
  if result is None:
    return
  return os.path.abspath(result).replace("\\", "/")


class NoLogger(Logger):
  def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=1):
    pass


class MS2:
  version = _module_info.version

  def __init__(self, *,
               logger: Logger = None,
               **kw):
    self.args = sys.argv
    self.encoding: str = "utf-8"
    self.env = os.environ
    self.import_code = "from MainShortcuts2 import ms"
    self.log: Logger = NoLogger("MainShortcuts2") if logger is None else logger
    self.MAIN_FILE: str | None = _get_main_file()
    self.MAIN_DIR: str | None = None if self.MAIN_FILE is None else os.path.dirname(self.MAIN_FILE)
    self.use_tmp_file = True
    self.reload()

  def reload(self):
    self.pid = os.getpid()

  @cached_property
  def credits(self) -> str:
    line = "------------------------------------------------"
    l = []
    l.append(line)
    l.append(("MainShortcuts2 %s" % self.version).center(len(line)))
    l.append("https://github.com/MainPlay-TG/MainShortcuts2.py")
    l.append(line)
    l.append("Разработчик: MainPlay TG")
    l.append("- Telegram - https://t.me/MainPlay_TG")
    l.append("- GitHub - https://github.com/MainPlay-TG")
    l.append("- Пожертвования - https://www.donationalerts.com/r/mainplay_tg (не знаю зачем, всё равно никто и копейки не даст)")
    l.append(line)
    l.append("Тестеры:")
    l.append("- MainPlay TG (https://t.me/MainPlay_TG)")
    l.append("- Кларк (https://t.me/Klark_Anthrofox)")
    l.append("- FOX (https://t.me/black_fox404)")
    l.append("- Ты (напиши мне в Telegram и в ближайшем обновлении тут будет твоя ссылка)")
    l.append(line)
    l.append("Спасибо за использование моей библиотеки".center(len(line)))
    return "\n".join(l)

  @cached_property
  def advanced(self):
    from . import advanced
    return advanced

  @cached_property
  def any2json(self):
    from . import any2json
    return any2json

  @cached_property
  def cfg(self):
    from .cfg import cfg
    return cfg

  @cached_property
  def dict(self):
    from . import dict
    return dict

  @cached_property
  def dir(self):
    from . import dir
    return dir

  @cached_property
  def file(self):
    from . import file
    return file

  @cached_property
  def json(self):
    from . import json
    return json

  @cached_property
  def list(self):
    from . import list
    return list

  @cached_property
  def ms2app(self):
    from . import ms2app
    return ms2app

  @cached_property
  def ms2hash(self):
    from . import ms2hash
    return ms2hash

  @cached_property
  def path(self):
    from . import path
    return path

  @cached_property
  def proc(self):
    from . import proc
    return proc

  @cached_property
  def regex(self):
    from . import regex
    return regex

  @cached_property
  def special_chars(self):
    from . import special_chars
    return special_chars

  @cached_property
  def str(self):
    from . import str
    return str

  @cached_property
  def term(self):
    from . import term
    return term

  @cached_property
  def types(self):
    from . import types
    return types

  @cached_property
  def utils(self):
    from . import utils
    return utils

  @cached_property
  def win(self):
    from . import win
    return win

  class ObjectBase(object):
    def __new__(cls, *a, **b):
      self = object.__new__(cls)
      return self

    def __init__(self, *a, **b):
      pass

    def __bool__(self):
      return True

    def __enter__(self):
      return self

    def __exit__(self, *a):
      if callable(getattr(self, "close", None)):
        self.close()

    def __repr__(self, args_string="..."):
      cls = type(self)
      return "%s.%s(%s)" % (cls.__module__, cls.__name__, args_string)

    @classmethod
    def _from_kw(cls, kw):
      return cls(**kw)

  @property
  def now(self) -> float:
    """Текущее локальное время (`timestamp`)"""
    return time()

  @property
  def now_dt(self) -> datetime:
    """Текущее локальное время (`datetime`)"""
    return datetime.fromtimestamp(time())

  @property
  def utcnow(self) -> float:
    """Текущее время по UTC (`timestamp`)"""
    return self.utcnow_dt.timestamp()
  if timezone is None:
    @property
    def utcnow_dt(self) -> datetime:
      """Текущее время по UTC (`datetime`)"""
      return datetime.utcfromtimestamp(time())
  else:
    @property
    def utcnow_dt(self) -> datetime:
      """Текущее время по UTC (`datetime`)"""
      return datetime.fromtimestamp(time(), timezone.utc)


ms = MS2()



"""Работа только с файлами"""
import builtins
import os
from .core import ms
from .path import PATH_TYPES, path2str
from .types import NotAFileError
from typing import *
from warnings import warn
TMP_SUFFIX = ".MS2_TMP"
CHUNK_SIZE = 64 * 1024


def _check(path, real=False, **kw) -> str:
  path = path2str(path, **kw)
  if os.path.exists(path):
    if not os.path.isfile(path):
      raise NotAFileError(path)
    if real:
      return os.path.realpath(path)
  return path


def _write(path, data, mode, mkdir=False, **kw):
  if "use_tmp_file" in kw:
    if not kw.pop("use_tmp_file"):
      # Записывать файл напрямую небезопасно, всегда использовать временный файл
      warn("Argument 'use_tmp_file' is deprecated, always use temporary file", DeprecationWarning)
  file = _check(path, True)
  if mkdir:
    os.makedirs(os.path.dirname(file), exist_ok=True)
  # Запись в временный файл
  tmp_file = file + TMP_SUFFIX
  with builtins.open(tmp_file, mode, **kw) as f:
    result = f.write(data)
  # Замена файла
  os.replace(tmp_file, file)
  return result


def read(path: PATH_TYPES, encoding: str = None, **kw) -> str:
  """Получить текст из файла"""
  kw["encoding"] = ms.encoding if encoding is None else encoding
  with builtins.open(_check(path), "r", **kw) as f:
    return f.read()


def write(path: PATH_TYPES, data: str, encoding: str = None, mkdir: bool = False, **kw) -> int:
  """Записать текст в файл"""
  kw.setdefault("newline", "\n")
  kw["encoding"] = ms.encoding if encoding is None else encoding
  return _write(path, data, "w", mkdir, **kw)


def load(path: PATH_TYPES, **kw) -> bytes:
  """Получить байты из файла"""
  with builtins.open(_check(path), "rb", **kw) as f:
    return f.read()


def save(path: PATH_TYPES, data: bytes, mkdir: bool = False, **kw) -> int:
  """Записать байты в файл"""
  return _write(path, data, "wb", mkdir, **kw)


def copy(path: PATH_TYPES, dest: PATH_TYPES, **kw):
  """Копировать файл"""
  return ms.path.copy(_check(path), dest, **kw)


def delete(path: PATH_TYPES, **kw):
  """Удалить файл"""
  return ms.path.delete(_check(path), **kw)


def in_dir(path: str, dir: str, **kw) -> bool:
  """Находится ли файл в указанной папке"""
  return ms.path.in_dir(_check(path), dir, **kw)


def link(path: PATH_TYPES, dest: PATH_TYPES, **kw):
  """Сделать символическую ссылку на файл"""
  return ms.path.link(_check(path), dest, **kw)


def move(path: PATH_TYPES, dest: PATH_TYPES, **kw):
  """Переместить файл"""
  return ms.path.move(_check(path), dest, **kw)


def rename(path: PATH_TYPES, name: PATH_TYPES, **kw):
  """Переименовать файл"""
  return ms.path.rename(_check(path), name, **kw)


def compare(path1: PATH_TYPES, path2: PATH_TYPES, method: str = "bin", chunk_size=CHUNK_SIZE) -> bool:
  """Одинаковые ли файлы (сравнение размера и содержимого)"""
  p1 = _check(path1)
  p2 = _check(path2)
  s1 = os.stat(p1)
  s2 = os.stat(p2)
  if os.path.samestat(s1, s2):
    return True
  if s1.st_size != s2.st_size:
    return False
  method = method.lower()
  if method in ["binary", "bin", "bytes"]:  # Сравнение по байтам
    with open(p1, "rb") as f1:
      with open(p2, "rb") as f2:
        while True:
          b1 = f1.read(chunk_size)
          b2 = f2.read(chunk_size)
          if b1 != b2:
            return False
          if not b1:
            break
    return True
  from .ms2hash import HASH_TYPES
  if method in HASH_TYPES:  # Сравнение контрольной суммы
    import hashlib
    with open(p1, "rb") as f1:
      hash1 = hashlib.new(method)
      while True:
        b1 = f1.read(chunk_size)
        if not b1:
          break
        hash1.update(b1)
    with open(p2, "rb") as f2:
      hash2 = hashlib.new(method)
      while True:
        b2 = f2.read(chunk_size)
        if not b2:
          break
        hash2.update(b2)
    return hash1.digest() == hash2.digest()
  raise ValueError("Unknown method: " + repr(method))



from datetime import timedelta
from functools import cached_property, total_ordering


def to_td(obj):
  if isinstance(obj, timedelta):
    return obj
  if isinstance(obj, (float, int)):
    return timedelta(seconds=obj)
  if isinstance(obj, dict):
    return timedelta(**obj)
  if isinstance(obj, TDInfo):
    return obj.timedelta
  raise TypeError(f"Unsupported type {type(obj)}")


@total_ordering
class TDInfo:
  _replace_keys = ("days", "hours", "microseconds", "minutes", "seconds")
  _letters = ["d", "h", "m", "s"]

  def __init__(self, td: timedelta):
    self._td = td
    self._override_letters = self._letters.copy()

  def __add__(self, other):
    return type(self)(self.timedelta + to_td(other))

  def __sub__(self, other):
    return type(self)(self.timedelta - to_td(other))

  def __eq__(self, other):
    if isinstance(other, TDInfo):
      return self.timedelta == other.timedelta
    return False

  def __lt__(self, other):
    if isinstance(other, TDInfo):
      return self.timedelta < other.timedelta
    return NotImplemented

  def __str__(self):
    l = self._override_letters
    parts = []
    if self.days:
      parts.append(f"{self.days}{l[0]}")
    if self.hours:
      parts.append(f"{self.hours}{l[1]}")
    if self.minutes:
      parts.append(f"{self.minutes}{l[2]}")
    if self.seconds:
      parts.append(f"{self.seconds}{l[3]}")
    if parts:
      return " ".join(parts)
    return f"0{l[3]}"

  @cached_property
  def days(self) -> int:
    """Кол-во дней (не ограничено)"""
    return self._td.days

  @cached_property
  def hours(self) -> int:
    """Кол-во часов (`range(24)`)"""
    return self._td.seconds // 3600

  @cached_property
  def minutes(self) -> int:
    """Кол-во минут (`range(60)`)"""
    return self._td.seconds // 60 % 60

  @cached_property
  def seconds(self) -> int:
    """Кол-во секунд (`range(60)`)"""
    return self._td.seconds % 60

  @cached_property
  def microseconds(self) -> int:
    """Кол-во микросекунд (`range(1000000)`)"""
    return self._td.microseconds

  @cached_property
  def timedelta(self) -> timedelta:
    """Оригинальный `timedelta`"""
    return self._td

  @cached_property
  def total_days(self) -> float:
    """Общее кол-во дней"""
    return self.total_seconds / 86400

  @cached_property
  def total_hours(self) -> float:
    """Общее кол-во часов"""
    return self.total_seconds / 3600

  @cached_property
  def total_minutes(self) -> float:
    """Общее кол-во минут"""
    return self.total_seconds / 60

  @cached_property
  def total_seconds(self) -> float:
    """Общее кол-во секунд"""
    return self._td.total_seconds()

  @classmethod
  def create(cls, days=0, seconds=0, microseconds=0, milliseconds=0, minutes=0, hours=0, weeks=0, **kw):
    """Создать с параметрами как у `timedelta`"""
    return cls(timedelta(days, seconds, microseconds, milliseconds, minutes, hours, weeks, **kw))

  def replace(self, **kw):
    """Создать с заменой параметров"""
    for k in self._replace_keys:
      if not k in kw:
        kw[k] = getattr(self, k)
    return type(self)(timedelta(**kw))



"""Различные утилиты, иногда требующие сторонних модулей"""
import atexit
import builtins
import importlib
import logging
import os
import sys
import time
from .core import ms
from functools import wraps
from typing import *
from warnings import warn
cache = {}


def return_False(*a, **b):
  """Вернуть False при любых аргументах"""
  return False


def return_None(*a, **b):
  """Вернуть None при любых аргументах"""
  return None


def return_True(*a, **b):
  """Вернуть True при любых аргументах"""
  return True


class MiddlewareBase:
  """middleware для функции"""

  def __init__(self, func):
    self._init(func)

  def _init(self, func):
    self._args = None
    self._exception = None
    self._kwargs = None
    self._result = None
    self._time = None
    self._traceback = None
    self.completed_with_exc: Union[None, bool] = None
    self.completed: bool = False
    self.func = func
    self.launched: bool = False
    if not hasattr(self, "ignore_exceptions"):
      self.ignore_exceptions: bool = False

  def _check_completed(self):
    if not self.completed:
      raise RuntimeError("The function has not yet been completed")

  def _check_launched(self):
    if not self.launched:
      raise RuntimeError("The function has not yet been launched")

  def _run(self, args, kwargs):
    from time import time
    if self.launched:
      raise RuntimeError("You can only start a function once once")
    self._args = args
    self._kwargs = kwargs
    self.launched = True
    try:
      self.before()
    except Exception:
      if not self.ignore_exceptions:
        raise
    started_at = time()
    try:
      self._result = self.func(*self.args, **self.kwargs)
      self.completed_with_exc = False
    except Exception as e:
      from traceback import format_exc
      self._exception = e
      self._traceback = format_exc()
      self.completed_with_exc = True
    self._time = time() - started_at
    self.completed = True
    try:
      self.after()
    except Exception:
      if not self.ignore_exceptions:
        raise
    return self.result

  @property
  def args(self) -> tuple:
    """Аргументы, переданные функции"""
    self._check_launched()
    return self._args

  @args.setter
  def args(self, value: Iterable):
    if value is None:
      value = []
    self._args = tuple(value)

  @property
  def exception(self) -> None | Exception:
    """Исключение, возникшее в процессе работы функции"""
    self._check_completed()
    return self._exception

  @property
  def kwargs(self) -> dict[str, Any]:
    """Именованные аргументы, переданные функции"""
    self._check_launched()
    return self._kwargs

  @kwargs.setter
  def kwargs(self, value: dict[str, Any]):
    if value is None:
      value = {}
    for k in value.keys():
      assert type(k) == str
    self._kwargs = value

  @property
  def result(self) -> Any:
    """Результат работы функции"""
    self._check_completed()
    if self.completed_with_exc:
      raise self.exception  # type: ignore
    return self._result

  @property
  def time(self) -> float:
    """Продолжительность работы функции (сек)"""
    self._check_completed()
    return self._time

  @property
  def traceback(self) -> None | str:
    self._check_completed()
    return self._traceback

  def before(self):
    """Перед запуском функции. Можно обработать/изменить аргументы"""
    pass

  def after(self):
    """После запуска функции. Можно обработать результат или исключение"""
    pass


def args2kwargs(func: Callable, args: Iterable = (), kwargs: dict[str, Any] = {}) -> tuple[tuple, dict[str, Any]]:
  """Преобразовать `args` в `kwargs` | `inspect`"""
  import inspect
  kw = kwargs.copy()
  args = list(args)
  spec = inspect.getfullargspec(func)
  for i in inspect.signature(func).parameters:
    if i != spec.varargs:
      if i != spec.varkw:
        if not i in kw:
          kw[i] = args.pop(0)
  if len(args) > 0:
    if spec.varargs != None:
      raise TypeError("Too many arguments")
  return tuple(args), kw


async def async_download_file(url: str, path: str, *, cb_end=return_None, cb_progress=return_None, cb_start=return_None, chunk_size: int = 1024, delete_on_error: bool = True, **kw) -> int:
  """Асинхронная функция для скачивания файла | `aiohttp`"""
  kw.setdefault("method", "GET")
  kw["url"] = url
  async with async_request(**kw) as resp:  # type: ignore
    if callable(getattr(path, "write", None)):
      f: IO[bytes] = path
    else:
      f = open(path, "wb")
    with f:
      size = 0
      await cb_start(f, resp, size)
      try:
        async for chunk in resp.content.iter_chunked(chunk_size):
          size += f.write(chunk)
          await cb_progress(f, resp, size)
      except:
        if delete_on_error:
          if os.path.isfile(path):
            os.remove(path)
        raise
      cb_end(f, resp, size)
  return size


async def async_request(method: str, url: str, *, ignore_status: bool = False, session=None, **kw):
  """Асинхронный HTTP запрос | `aiohttp`"""
  from importlib import import_module
  aiohttp = import_module("aiohttp")
  # import aiohttp
  resp = None
  if isinstance(method, aiohttp.ClientResponse):
    resp = method
  if isinstance(url, aiohttp.ClientResponse):
    if not resp is None:
      raise TypeError("Only one argument can be `Response`")
    resp = url
  if resp is None:
    if session is None:
      session = aiohttp.ClientSession()
    kw["method"] = method
    kw["url"] = url
    resp = await session.request(**kw)
  if not ignore_status:
    resp.raise_for_status()
  return resp


def async2sync(func: Callable) -> Callable:
  """Превратить асинхронную функцию в синхронную | `asyncio`, `concurrent`"""
  import asyncio
  import concurrent.futures
  pool = concurrent.futures.ThreadPoolExecutor()

  def wrapper(*args, **kwargs):
    return pool.submit(asyncio.run, func(*args, **kwargs)).result()
  return wrapper


def get_my_ip() -> str:
  """Получить глобальный IP | `requests`"""
  with sync_request("GET", "https://api.ipify.org?format=json") as resp:
    ip = resp.json()["ip"]
  return ip


def is_async(func: Callable) -> bool:
  """Является ли функция асинхронной | `inspect`"""
  import inspect
  return inspect.iscoroutinefunction(func)


def is_sync(func: Callable) -> bool:
  """Является ли функция синхронной | `inspect`"""
  return not is_async(func)


def middleware(cls: type[MiddlewareBase]):
  """middleware для функции в виде декоратора"""
  def decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
      return cls(func)._run(args, kwargs)
    return wrapper
  return decorator


def randfloat(min: float, max: float = None) -> float:
  """Случайное число | `random`"""
  from random import random
  if max == None:
    max = min
    min = 0
  return min + (random() * (max - min))


def randstr(length: int, symbols: str = "0123456789abcdefghijklmnopqrstuvwxyz") -> str:
  """Случайная строка | `random`"""
  from random import choice
  t = ""
  while len(t) < length:
    t += choice(symbols)
  return t


def riop(**p_kw):
  """Запустить функцию в отдельном процессе | `multiprocessing`"""
  import multiprocessing
  p_kw.setdefault("daemon", False)

  def decorator(func):
    p_kw["target"] = func

    @wraps(func)
    def wrapper(*args, **kwargs) -> multiprocessing.Process:
      p_kw["args"] = args
      p_kw["kwargs"] = kwargs
      p = multiprocessing.Process(**p_kw)
      p.start()
      return p
    return wrapper
  return decorator


def riot(**t_kw):
  """Запустить функцию в отдельном потоке | `threading`"""
  import threading
  t_kw.setdefault("daemon", False)

  def decorator(func):
    t_kw["target"] = func

    @wraps(func)
    def wrapper(*args, **kwargs) -> threading.Thread:
      t_kw["args"] = args
      t_kw["kwargs"] = kwargs
      t = threading.Thread(**t_kw)
      t.start()
      return t
    return wrapper
  return decorator


def sync_download_file(url: str, path: str, *, cb_end=return_None, cb_progress=return_None, cb_start=return_None, chunk_size: int = 1024, delete_on_error: bool = True, **kw) -> int:
  """Синхронная функция для скачивания файла | `requests`"""
  kw.setdefault("method", "GET")
  kw["stream"] = True
  kw["url"] = url
  with sync_request(**kw) as resp:
    if callable(getattr(path, "write", None)):
      f: IO[bytes] = path
    else:
      f = open(path, "wb")
    with f:
      size = 0
      cb_start(f, resp, size)
      try:
        for chunk in resp.iter_content(chunk_size):
          size += f.write(chunk)
          cb_progress(f, resp, size)
      except:
        f.close()
        if delete_on_error:
          if os.path.isfile(path):
            os.remove(path)
        raise
      cb_end(f, resp, size)
  return size


download_file = sync_download_file


def sync_request(method: str, url: str, *, ignore_status: bool = False, session=None, **kw):
  """Синхронный HTTP запрос | `requests`"""
  try:
    import requests
  except ImportError as err:
    try:
      from pip._vendor import requests
    except ImportError:
      raise err
  resp = None
  if isinstance(method, requests.Response):
    resp = method
  if isinstance(url, requests.Response):
    if not resp is None:
      raise TypeError("Only one argument can be `Response`")
    resp = url
  if resp is None:
    if session is None:
      session = requests.Session()
    kw["method"] = method
    kw["url"] = url
    resp = session.request(**kw)
  if not ignore_status:
    resp.raise_for_status()
  return resp


request = sync_request


def sync2async(func: Callable) -> Callable:
  """Превратить синхронную функцию в асинхронную"""
  async def wrapper(*args, **kwargs):
    return func(*args, **kwargs)
  return wrapper


def uuid(format=None, *args, **kwargs) -> str:
  """Сгенерировать UUID | `uuid`"""
  import uuid
  if format is None:
    cls = uuid.uuid4
  else:
    cls = getattr(uuid, f"uuid{format}")
  return str(cls(*args, **kwargs))


def timedelta(time: Union[int, float, dict]):
  """Превратить число/словарь в `timedelta` | `datetime`"""
  import datetime
  if isinstance(time, datetime.timedelta):
    return time
  if isinstance(time, dict):
    return datetime.timedelta(**time)
  return datetime.timedelta(seconds=time)


def shebang_code(code: str, *, exe_name: Union[None, str] = None, exe_path: Union[None, str] = None, none_if_no_changes: bool = False, use_env: bool = True) -> Union[None, str]:
  """Вставить/заменить шебанг в коде. Если указать имя интерпретатора, путь будет найден с помощью `shutil.which`"""
  if (exe_name is None) and (exe_path is None):
    raise TypeError("Specify exe_name or exe_path")
  if not exe_name is None:
    if not exe_path is None:
      raise TypeError("exe_name and exe_path cannot be used together")
  if exe_path is None:
    if use_env:
      exe_path = "/bin/env " + exe_name
    else:
      import shutil
      exe_path = shutil.which(exe_name)
      if exe_path is None:
        raise Exception("Command %s not found in PATH" % exe_name)
  lines = code.split("\n")
  if none_if_no_changes:
    if lines[0] == "#!" + exe_path:
      return None
  while len(lines[0].strip()) == 0 or lines[0].startswith("#!"):
    del lines[0]
  return "#!" + exe_path + "\n" + "\n".join(lines)


def shebang_file(path: str, **kw) -> int:
  """Вставить/заменить шебанг в файле кода"""
  kw["code"] = ms.file.read(path)
  kw["none_if_no_changes"] = True
  result = shebang_code(**kw)
  if result is None:
    return 0
  return ms.file.write(path, result)


def remove_ANSI(text: str) -> str:
  """Убрать ANSI коды из текста | `re`"""
  cache_id = "remove_ANSI", 0
  if not cache_id in cache:
    import re
    cache[cache_id] = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
  return cache[cache_id].sub("", text)


OnlyOneInstanceError = ms.types.OnlyOneInstanceError


class OnlyOneInstance:
  """Запретить запуск одной программы несколько раз | `tempfile`, `fcntl`"""
  _win = sys.platform == "win32"

  def __init__(self, name: str = "main", lock_path: str = None):
    import tempfile
    self.name: str = name
    self.running = False
    if lock_path is None:
      lock_path = tempfile.gettempdir() + "/" + ms.MAIN_FILE.replace(":", "").replace("/", "_") + "." + name + ".lock"
    self.lock = ms.path.Path(lock_path, use_cache=False)
    if self._win:
      flags = os.O_CREAT | os.O_EXCL | os.O_RDWR

      def _enter():
        try:
          if os.path.exists(self.lock.path):
            os.unlink(self.lock.path)
          self.fd = os.open(self.lock.path, flags)
        except OSError as err:
          if err.errno == 13:
            raise OnlyOneInstanceError()
          raise

      def _exit():
        os.close(self.fd)
        os.unlink(self.lock.path)
    else:
      import fcntl
      flags = fcntl.LOCK_EX | fcntl.LOCK_NB

      def _enter():
        self.fp = open(self.lock.path, "w")
        self.fp.flush()
        try:
          fcntl.lockf(self.fp, flags)
        except IOError:
          raise OnlyOneInstanceError()

      def _exit():
        fcntl.lockf(self.fp, fcntl.LOCK_UN)
        self.fp.close()
        if os.path.exists(self.lock.path):
          os.unlink(self.lock.path)
    self._enter = _enter
    self._exit = _exit

  def __enter__(self):
    if self.running:
      raise OnlyOneInstanceError()
    self._enter()
    self.running = True
    try:
      self.on_enter()
    except Exception:
      pass
    return self

  def __exit__(self, a, b, c):
    if not self.running:
      return
    self._exit()
    self.running = False
    self.lock.delete()
    self.on_exit()

  @classmethod
  def wrap_func(cls, **kw):
    def deco(func):
      ooi = cls(**kw)

      def wrapper(*args, **kwargs):
        with ooi:
          return func(*args, **kwargs)
      return wrapper
    return deco

  def on_enter(self):
    pass

  def on_exit(self):
    pass


def multi_and(*values: bool) -> bool:
  for i in values:
    if not i:
      return False
  return True


def multi_or(*values: bool) -> bool:
  for i in values:
    if i:
      return True
  return False


def is_int(value: float) -> bool:
  return value == int(value)


def get_self_module(__name__: str):
  return sys.modules[__name__]


def check_programs(*progs: str, raise_error: bool = True) -> list[str]:
  """Проверить наличие программ в `$PATH` | `shutil`"""
  from shutil import which
  failed = []
  for i in progs:
    if which(i) is None:
      if not i in failed:
        failed.append(i)
  failed.sort()
  if raise_error:
    if len(failed) > 0:
      if len(failed) == 1:
        raise OSError("Failed to find program " + failed[0] + " in $PATH")
      raise OSError("Failed to find programs " + (", ".join(failed)) + " in $PATH")
  return failed


def handle_exception(on_exception=return_None, reraise: bool = True, exc_type: type[BaseException] = Exception):
  """Обернуть функцию в `try`/`except`"""
  def deco(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
      try:
        return func(*args, **kwargs)
      except exc_type as exc:
        on_exception(exc)
        if reraise:
          raise
    return wrapper
  return deco


def disable_warnings():
  """Отключить модуль `warnings`"""
  import warnings
  warnings.warn = return_None


def is_instance_of_one(obj, *classes: type) -> bool:
  """Это экземпляр одного из классов?"""
  for i in classes:
    if isinstance(obj, i):
      return True
  return False


def is_instance_of_all(obj, *classes: type) -> bool:
  """Это экземпляр всех классов?"""
  for i in classes:
    if not isinstance(obj, i):
      return False
  return True


def restore_deprecated(force: bool = False):
  """Восстановить устаревшие функции"""
  restored = []

  def setmethod(cls, name):
    def deco(func):
      if force or not hasattr(cls, name):
        try:
          setattr(cls, name, func)
          restored.append((cls, name, func))
        except Exception:
          pass
      return func
    return deco
  if "datetime" in sys.modules:
    from datetime import datetime

    @setmethod(datetime, "utcnow")
    def utcnow():
      return datetime.now(datetime.UTC)
  if "locale" in sys.modules:
    import locale

    @setmethod(locale, "resetlocale")
    def resetlocale(category=locale.LC_ALL):
      locale.setlocale(category, "")
  return restored


def _run_async_coro(coro, mode="run", **kw):
  import asyncio
  if mode == "loop":
    loop = asyncio.get_running_loop()
    return loop.run_until_complete(coro)
  if mode == "run":
    return asyncio.run(coro)
  raise ValueError("Unsupported mode: %s" % mode)


def main_func(_name_: str, exit: bool = True, async_mode="run"):
  import inspect

  def deco(func):
    if _name_ == "__main__":
      try:
        if inspect.iscoroutinefunction(func):
          result = _run_async_coro(func(), async_mode)
        else:
          result = func()
      except KeyboardInterrupt:
        import signal
        result = signal.SIGINT.value
      if exit:
        if isinstance(result, int):
          sys.exit(result)
        sys.exit()
    return func
  return deco


class decorators:
  def __init__(self):
    raise NotImplementedError()

  @staticmethod
  def append(obj: list):
    """`obj.append(func)`"""
    def deco(func):
      obj.append(func)
      return func
    return deco

  @staticmethod
  def setattr(obj: type, name):
    """`setattr(obj,name,func)`"""
    def deco(func):
      setattr(obj, name, func)
      return func
    return deco

  @staticmethod
  def setitem(obj: dict | list, index):
    """`obj[index]=func`"""
    def deco(func):
      obj[index] = func
      return func
    return deco


def fassert(value: bool, text: str = None):
  """`assert` независимо от `__debug__`"""
  if value:
    return value
  if text is None:
    raise AssertionError()
  raise AssertionError(text)


def add2pythonpath(dir: str, to_begin: bool = False):
  """Добавить папку в `sys.path`, если её там нет"""
  dir = ms.path.path2str(dir, to_abs=True)
  if not dir in sys.path:
    if to_begin:
      sys.path.insert(0, dir)
    else:
      sys.path.append(dir)


def run_pip(*args: str, internal: bool = False, **kw):
  """Запустить `pip`. Встроенный запуск может завершить работу всей программы! | `pip`"""
  if getattr(sys, "frozen", False):
    raise RuntimeError("The environment is frozen")
  from pip._internal.exceptions import PipError
  if internal:
    from pip._internal.cli.main_parser import parse_command
    from pip._internal.commands import create_command
    cmd_name, cmd_args = parse_command(args)
    kw.setdefault("isolated", "--isolated" in cmd_args)
    command = create_command(cmd_name, **kw)
    if command.main(cmd_args):
      raise PipError()
  else:
    import subprocess
    kw["args"] = [sys.executable, "-m", "pip"] + list(args)
    if subprocess.call(**kw):
      raise PipError()


def check_modules(*modules: str, _c: list[str] = None, _m: list[str] = None) -> list[str]:
  """Проверить наличие модулей. Не проверяет возможность импорта. Возвращает список отсутствующих модулей"""
  import pkg_resources
  checked = [] if _c is None else _c
  missing = [] if _m is None else _m
  for i in modules:
    req = i if isinstance(i, pkg_resources.Requirement) else pkg_resources.Requirement.parse(i)
    if not req.name in checked:
      try:
        dist = pkg_resources.get_distribution(req)
        checked.append(req.name)
        check_modules(*dist.requires(), _c=checked, _m=missing)
      except pkg_resources.DistributionNotFound:
        req_str = str(req)
        if not req_str in missing:
          missing.append(req_str)
  return missing


def auto_install_modules(*modules, print: bool | str = False, **pip_kw):
  """Автоматически установить недостающие модули | `pip`, `pkg_resources`"""
  # Проверить возможность установки
  import pkg_resources
  from pip._internal.exceptions import PipError
  if getattr(sys, "frozen", False):
    raise RuntimeError("The environment is frozen")
  _ = PipError, pkg_resources  # Убрать предупреждение "модуль не используется"
  missing = check_modules(*modules)
  if len(missing) > 0:
    if print:
      text = print if isinstance(print, str) else "Выполняется установка модулей %(modules)s, подождите немного"
      builtins.print(text % {"modules": ", ".join(missing)}, file=sys.stderr)
    args = ["install", "-U"] + missing
    run_pip(*args, **pip_kw)


def http_check_range_support(url: str, **kw) -> bool:
  """Проверить возможность скачать файл по частям"""
  kw.setdefault("headers", {})
  kw["headers"]["Range"] = "bytes=0-1"
  kw["method"] = "HEAD"
  kw["stream"] = True
  kw["url"] = url
  with sync_request(**kw) as resp:
    if resp.status_code == 206:
      return True
    if "bytes" in resp.headers.get("Accept-Ranges", "").lower():
      return True
    return False


class MultiContext:
  def __init__(self, *objects):
    self.enter_handlers = []
    self.exit_handlers = []
    self.suppress_exc: None | type[BaseException] | set[type[BaseException]] = None
    self.add_objects(*objects)

  def __enter__(self):
    for i in self.enter_handlers:
      i()
    return self

  def __exit__(self, etype, exc, etb):
    for i in self.exit_handlers:
      try:
        i(etype, exc, etb)
      except Exception:
        pass
    if not self.suppress_exc is None:
      if isinstance(self.suppress_exc, (list, set)):
        self.suppress_exc = tuple(self.suppress_exc)
      if isinstance(exc, self.suppress_exc):
        return True

  def _cl(self, l: list):
    if len(l) == 1:
      if isinstance(l[0], (list, set, tuple)):
        return list(l[0])
    return l

  def add_enter_handlers(self, *handlers):
    for i in self._cl(handlers):
      self.enter_handlers.append(i)

  def add_exit_handlers(self, *handlers):
    for i in self._cl(handlers):
      self.exit_handlers.append(i)

  def add_objects(self, *objects):
    """Добавить один/несколько объектов в контекст"""
    for i in self._cl(objects):
      self.enter_handlers.append(getattr(i, "__enter__", return_None))
      self.exit_handlers.append(getattr(i, "__exit__", return_None))

  def add_obj(self, obj):
    """Добавить объект в контекст и вернуть этот объект"""
    self.add_objects(obj)
    return obj


def generator2list(list=builtins.list):
  """Превратить функцию с `yield` в функцию, которая возвращает список"""
  def deco(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
      return list(func(*args, **kwargs))
    return wrapper
  return deco


def mini_log(msg: str, *values, **kw):
  """Напечатать текст в `stderr` с форматированием через `%`"""
  kw.setdefault("file", sys.stderr)
  if len(values) == 0:
    return print(msg, **kw)
  if len(values) == 1:
    values = values[0]
  print(msg % values, **kw)


def print_stderr(*values, **kw):
  """Тот же `print`, но всегда в `stderr`"""
  kw["file"] = sys.stderr
  print(*values, **kw)


def setattr_if_not_exists(obj, name: str, value, if_None=False):
  """Установить атрибут, если его нет"""
  if ((getattr(obj, name, None) is None) if if_None else (not hasattr(obj, name))):
    setattr(obj, name, value)


def call(func, args=[], kwargs={}):
  """Вызвать функцию. Если поставить как декоратор, функция будет вызвана сразу после определения и вместо неё будет сохранён результат её работы"""
  return func(*args, **kwargs)


class TempFlag(ms.types.BoolFlag):
  def __init__(self, value=False):
    super().__init__(value)
    warn("Use `ms.types.BoolFlag` instead", DeprecationWarning, stacklevel=2)

  @property
  def state(self):
    return self.value

  @state.setter
  def state(self, v):
    self.value = bool(v)


class SimpleLogger(logging.Logger):
  """Простой логгер"""
  CRITICAL: int = logging.CRITICAL
  DEBUG: int = logging.DEBUG
  ERROR: int = logging.ERROR
  FATAL: int = logging.FATAL
  INFO: int = logging.INFO
  NOTSET: int = logging.NOTSET
  WARN: int = logging.WARN
  WARNING: int = logging.WARNING

  def __init__(self, name: str, level: int = logging.INFO, full_global=False, init_msg=True):
    # Выбрать случайное название
    if name is None:
      name = "UnnamedLogger_" + uuid()
    logging.Logger.__init__(self, name, level)
    atexit.register(self.close)
    self.formatter = logging.Formatter("%(asctime)s %(levelname)s: %(message)s")
    self.logged_levels: set[int] = set()
    # Ллогирование в stderr
    self.addHandler(logging.StreamHandler(sys.stderr))
    self.update_handlers()
    # Глобальный логгер
    if full_global:
      self.set_as_global()
      self.set_excepthook()
    # Первое сообщение
    if init_msg:
      self._init_msg()

  def _init_msg(self):
    now = ms.utcnow_dt
    level_name = logging._levelToName.get(self.level, "UNNAMED")
    self.info("Date: %s UTC", now.strftime("%d.%m.%Y %H:%M:%S"))
    self.info("Platform: %s", sys.platform)
    self.info("Python: %s", sys.version)
    self.info("MainShortcuts2: %s", ms.version)
    self.info("Logger: %s", self.name)
    self.info("Log level: %s (%s)", (level_name, self.level))

  def _log(self, level, *args, **kw):
    if self.disabled:
      return
    self.logged_levels.add(level)
    logging.Logger._log(self, level, *args, **kw)

  def _handle_exc(self, etype, evalue, etrace):
    """Для `sys.excepthook`"""
    self.exception("Unhandled exception", exc_info=(etype, evalue, etrace))

  @property
  def max_level(self) -> int:
    """Максимальный уровень, который был выведен в лог"""
    if len(self.logged_levels) == 0:
      return 0
    return max(self.logged_levels)

  def conf_from_env(self, all=True):
    """Настроить логгер из переменных окружения (`DEBUG` и `NO_LOG`)"""
    if os.environ.get("NO_LOG", False):
      self.disabled = True
    if os.environ.get("DEBUG", False):
      if all:
        self.set_all_levels(self.DEBUG)
      else:
        self.setLevel(self.DEBUG)

  def set_all_levels(self, level: int):
    """Установить уровень для себя и всех обработчиков"""
    self.setLevel(level)
    for i in self.handlers:
      i.level = self.level

  def set_as_global(self):
    """Установить логгер как глобальный (`MainShortcuts2` и все, кто использует `sys.logger`)"""
    ms.log = self
    sys.logger = self

  def set_excepthook(self):
    """Заменить `sys.excepthook` на этот логгер"""
    sys.excepthook = self._handle_exc

  def set_for_module(self, *module_names: str, autoimport=True, var_names: set[str] = ["logger"]):
    """Установить логгер основным для модулей"""
    # Импорт модулей, если они ещё не импортированы
    if autoimport:
      for i in module_names:
        importlib.import_module(i)
    for i in module_names:
      # Только если модуль импортирован
      if i in sys.modules:
        # Установить логгер для указанных переменных
        for j in var_names:
          setattr(sys.modules[i], j, self)

  def update_handlers(self, level=None, formatter=None):
    """Установить уровень и форматтер всем обработчикам"""
    if formatter is None:
      formatter = self.formatter
    if level is None:
      level = self.level
    for i in self.handlers:
      i.setFormatter(formatter)
      i.setLevel(level)

  def close(self):
    """Закрыть логгер и все обработчики"""
    for i in self.handlers:
      if hasattr(i, "close"):
        if callable(i.close):
          i.close()


class DirLogger(SimpleLogger):
  """Логгер с записью в папку. Название файла содержит дату и время создания лога"""

  def __init__(self, name: str, dir: str, latest_log=True, **kw):
    init_msg = kw.get("init_msg", True)
    kw["init_msg"] = False
    SimpleLogger.__init__(self, name, **kw)
    self.remove_if_no_records_for_level = 0
    # Создание папки
    self.dir = ms.path.Path(dir)
    ms.dir.create(self.dir)
    # Файловый обработчик
    file_path = self.dir + "/%s.log" % ms.utcnow_dt.strftime("%Y-%m-%d_%H-%M-%S")
    self.file_handler = logging.FileHandler(file_path, encoding=ms.encoding)
    self.file_handler.formatter = self.formatter
    self.file_handler.level = self.level
    self.addHandler(self.file_handler)
    # Первое сообщение
    if init_msg:
      self._init_msg()
    # Символическая ссылка на последний лог
    if latest_log:
      try:
        ms.path.delete(self.dir + "/latest.log")
        ms.file.link(file_path, self.dir + "/latest.log", force=True)
      except Exception as exc:
        self.warning("Failed to create link 'latest.log'", exc_info=exc)

  @property
  def file_path(self):
    return self.file_handler.baseFilename

  def close(self):
    SimpleLogger.close(self)
    # Удалить файл, если не было важных записей
    if self.remove_if_no_records_for_level > self.max_level:
      ms.file.delete(self.file_path)


def sleep(seconds: float):
  """Ожидать N секунд"""
  if not isinstance(seconds, (float, int)):
    if hasattr(seconds, "total_seconds"):
      if callable(seconds.total_seconds):
        seconds = seconds.total_seconds()
      else:
        if isinstance(seconds.total_seconds, (float, int)):
          seconds = seconds.total_seconds
  if seconds > 0:
    time.sleep(seconds)


def which_real(cmd: str, **kw) -> str | None:
  """Получить реальный путь к исполняемому файлу"""
  from shutil import which
  result = which(cmd, **kw)
  if result:
    return os.path.realpath(result)


def iter_caller_modules():
  import inspect
  frame = inspect.currentframe()
  try:
    frame = frame.f_back
    while frame is not None:
      caller_module = inspect.getmodule(frame)
      if caller_module:
        yield caller_module
      frame = frame.f_back
  finally:
    del frame



from ._sql_base import *
from MainShortcuts2 import ms
from MainShortcuts2.ex import sqlite_ex
from typing import TYPE_CHECKING
__all__ = ["check_type", "Database", "ObjectBase"]


def _gws(where: dict, sep: str = " AND "):
  return sep.join(i + "=?" for i in where)


class Database(DatabaseBase):
  """SQLite3"""
  MEMORY = ":memory:"

  def __init__(self, path: str, **kw):
    if path == self.MEMORY:
      kw.setdefault("connect_on_init", False)
      kw["autosave"] = False
    else:
      ms.dir.create(ms.path.Path(path).parent_dir)
    kw["database"] = path
    DatabaseBase.__init__(self, **kw)
  if TYPE_CHECKING:
    def cursor(self) -> sqlite_ex.Cursor:
      return super().cursor()

  @classmethod
  def create_in_memory(cls, **kw):
    """Создать базу данных в оперативной памяти"""
    kw["path"] = cls.MEMORY
    return cls(**kw)

  def _update_schema(self, schema):
    with self.cursor() as cur:
      for table in schema:
        columns_str = ",".join(k + " " + v for k, v in schema[table].items())
        cur.execute(f"CREATE TABLE IF NOT EXISTS {table} ({columns_str});")
        cur.execute(f"PRAGMA table_info({table});")
        exist_cols = [i[1] for i in cur.fetchall()]
        for col_name, col_type in schema[table].items():
          if not col_name in exist_cols:
            cur.execute(f"ALTER TABLE {table} ADD COLUMN {col_name} {col_type};")
    self.save()

  def _connect(self):
    self.conn = sqlite_ex.connect(**self.conn_kw)

  def delete(self, table: str, where: dict):
    """Удалить строки из таблицы"""
    if isinstance(table, ObjectBase):
      table = table._table
    self.exec(*sqlite_ex.make_delete_sql(table, where), fetch=False)

  def insert(self, table: str, values: dict):
    """Вставить новую строку в таблицу"""
    if isinstance(table, ObjectBase):
      table = table._table
    self.exec(*sqlite_ex.make_insert_sql(table, values), fetch=False)

  def select(self, table: str, columns: str | list[str], where: dict) -> list[tuple]:
    """Выбрать строки из таблицы"""
    if isinstance(table, ObjectBase):
      table = table._table
    if not isinstance(columns, str):
      columns = ",".join(columns)
    if where:
      return self.exec(*sqlite_ex.make_select_sql(table, columns, where), fetch=True)
    return self.exec(f"SELECT {columns} FROM {table};", fetch=True)

  def update(self, table: str, values: dict, where: dict):
    """Изменить строки в таблице"""
    if isinstance(table, ObjectBase):
      table = table._table
    self.exec(*sqlite_ex.make_update_sql(table, values, where), fetch=False)

  def save(self):
    if self.connected:
      if self.conn_kw["database"] != self.MEMORY:
        self.conn.commit()

  def select_adv(self, table: str, columns: str | list[str], where: dict = None, order_by: str = None, limit: int = None, offset: int = None, other: str = None):
    if isinstance(table, ObjectBase):
      table = table._table
    if not isinstance(columns, str):
      columns = ",".join(columns)
    code = f"SELECT {columns} FROM {table}"
    if where:
      if isinstance(where, str):
        code += f" WHERE {where}"
      else:
        code += f" WHERE {_gws(where)}"
    if order_by:
      code += f" ORDER BY {order_by}"
    if limit:
      code += f" LIMIT {limit}"
    if offset:
      code += f" OFFSET {offset}"
    if other:
      code += f" {other}"
    if where:
      return self.exec(code, where.values(), fetch=True)
    return self.exec(code, fetch=True)



import sys
from argparse import ArgumentParser
from logging import getLogger, Logger
from MainShortcuts2 import ms
from MainShortcuts2.utils import mini_log
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *
S_B = 1
S_KB = S_B
S_MB = S_KB * 1024
S_GB = S_MB * 1024
S_TB = S_GB * 1024


def humanize_filesize(size: int):
  if S_KB > size:
    return f"{size} B"
  if S_MB > size:
    return f"{size / S_KB:.2f} KB"
  if S_GB > size:
    return f"{size / S_MB:.2f} MB"
  if S_TB > size:
    return f"{size / S_GB:.2f} GB"
  return f"{size / S_TB:.2f} TB"


class QApplication(QtWidgets.QApplication):
  def __init__(self, *a, **kw):
    super().__init__(*a, **kw)
    self.logger = getLogger(type(self).__name__)

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    if exc_val is not None:
      self.logger.exception("Unhandled exception", exc_info=exc_val)


def main_func(_name_):
  def deco(func):
    if _name_ != "__main__":
      return func

    def wrapper(*args, **kwargs):
      with QApplication(sys.argv) as app:
        kwargs["qapp"] = app
        func(*args, **kwargs)
        sys.exit(app.exec())
    return wrapper
  return deco



import math
import os
from collections import Counter
from functools import cached_property
from io import BytesIO
from PIL import Image as ImageModule
from PIL import ImageDraw
from PIL.Image import Resampling
from typing import TYPE_CHECKING, Self
COLOR_TYPES = float | tuple[int, ...] | None


def is_image_file(name: str) -> bool:
  ext = os.path.splitext(name)[1].lower()
  return ext in ImageModule.registered_extensions()


try:
  import numpy as np
  # NumPy быстрее

  def _replace_img_color(img: "Image", old_color, new_color):
    arr = np.array(img)
    if img.mode in ("L", "P"):
      mask = (arr == old_color)
    else:
      mask = np.all(arr == np.array(old_color), axis=-1)
    arr[mask] = new_color
    result = Image.from_standard(ImageModule.fromarray(arr))
    assert img.size == result.size
    img._im = result._im
    img._mode = result._mode
except Exception:
  def _replace_img_color(img: "Image", old_color, new_color):
    for coord, color in img.iter_pixels():
      if color == old_color:
        img.putpixel(coord, new_color)


class Image(ImageModule.Image):
  def __getitem__(self, key) -> COLOR_TYPES:
    return self.getpixel(key)

  def __setitem__(self, key, value):
    self.putpixel(key, value)

  def _new(self, im):
    return self.from_standard(super()._new(im))

  @cached_property
  def aspect_ratio(self) -> tuple[int, int]:
    """Соотношение сторон"""
    cd = math.gcd(*self.size)
    return self.width // cd, self.height // cd

  @cached_property
  def aspect_ratio_float(self) -> float:
    """Соотношение сторон (в виде `float`)"""
    return self.width / self.height

  @cached_property
  def converter(self):
    """Конвертер изображения в объекты других библиотек"""
    return Converter(self)

  @cached_property
  def draw(self) -> ImageDraw.ImageDraw:
    """Рисовалка"""
    return ImageDraw.Draw(self)

  @classmethod
  def from_standard(cls, other: ImageModule.Image):
    """Преобразовать стандартный `PIL.Image.Image` в текущий класс"""
    other.load()
    self = cls()
    self._exif = other._exif
    self._im = other.im
    self._mode = other.mode
    self._size = other.size
    self.info = None if other.info is None else other.info.copy()
    self.palette = None if other.palette is None else other.palette.copy()
    self.readonly = other.readonly
    return self

  @classmethod
  def open(cls, fp, mode="r", formats=None):
    """Загрузить из файла или `BytesIO`"""
    other = ImageModule.open(fp, mode, formats)
    return cls.from_standard(other)

  @classmethod
  def new(cls, mode, size, color=0):
    other = ImageModule.new(mode, size, color)
    return cls.from_standard(other)

  def copy(self):
    return self.from_standard(super().copy())

  def to_bytesio(self, format: str, **kw) -> BytesIO:
    """Сохранить в `BytesIO`"""
    f = BytesIO()
    self.save(f, format, **kw)
    f.seek(0)
    return f

  def iter_coords(self):
    """Итерация координат сверху вниз слева направо"""
    for y in range(self.height):
      for x in range(self.width):
        yield x, y

  def iter_pixels(self):
    """Итерация координат и значений пикселей"""
    for coord in self.iter_coords():
      yield coord, self.getpixel(coord)

  def replace_color(self, old_color, new_color):
    """Заменить один цвет на другой"""
    if old_color == new_color:
      return
    if isinstance(old_color, (list, tuple)) and isinstance(new_color, (list, tuple)):
      if len(old_color) != len(new_color):
        raise ValueError("The lengths of the colors are different")
    _replace_img_color(self, old_color, new_color)

  def count_colors(self):
    """Посчитать количество цветов"""
    return Counter(self.getdata())

  def scale(self, scale: float, **kw):
    """Умножить размер изображения"""
    new_size = int(self.width * scale), int(self.height * scale)
    return self.from_standard(self.resize(new_size, **kw))

  def to_dict_matrix(self):
    """Преобразовать в матрицу (словарь)"""
    return {coord: color for coord, color in self.iter_pixels()}

  def to_list_matrix(self):
    """Преобразовать в матрицу (вложенный список)"""
    return [[self.getpixel((x, y)) for x in range(self.width)] for y in range(self.height)]

  def optimize(self, add_methods: list = None):
    """Оптимизировать изображение (если возможно). Возвращает список использованных методов"""
    methods = []
    if add_methods:
      methods.extend(add_methods)
    if self.mode.upper() in ("LA", "PA", "RGBA"):
      # Если есть прозрачность
      n = len(self.mode) - 1
      p = tuple([0] * len(self.mode))

      def clear_invisible_colors(coords, color):
        # И пиксель полностью прозрачный
        if color[n] == 0:
          # Удалить его цвет для лучшего сжатия при сохранении
          return p
      methods.append(clear_invisible_colors)
    if methods:
      for coords, color in self.iter_pixels():
        edited = False
        for method in methods:
          new_color = method(coords, color)
          if new_color is not None:
            color = new_color
            edited = True
        if edited:
          self.putpixel(coords, color)
    return methods
  if TYPE_CHECKING:
    def resize(self, size: tuple[int, int] | list[int] | ImageModule.NumpyArray, resample: int | None = None, box: tuple[float, float, float, float] | None = None, reducing_gap: float | None = None) -> Self:
      return super().resize(size, resample, box, reducing_gap)

    def reduce(self, factor: int | tuple[int, int], box: tuple[int, int, int, int] | None = None) -> Self:
      return super().reduce(factor, box)

    def rotate(self, angle: float, resample: Resampling = Resampling.NEAREST, expand: int | bool = False, center: tuple[float, float] | None = None, translate: tuple[int, int] | None = None, fillcolor: float | tuple[float, ...] | str | None = None) -> Self:
      return super().rotate(angle, resample, expand, center, translate, fillcolor)


class Converter:
  def __init__(self, img: Image):
    self._img = img

  @property
  def img(self):
    if self._img.mode == "RGBA":
      return self._img
    return self._img.convert("RGBA")
  # PySide6

  def to_ps6_icon(self):
    """`PySide6.QtGui.QIcon`"""
    from PySide6.QtGui import QIcon
    return QIcon(self.to_ps6_pixmap())

  def to_ps6_image(self):
    """`PySide6.QtGui.QImage`"""
    from PySide6.QtGui import QImage
    img = self.img
    return QImage(img.tobytes(), img.width, img.height, QImage.Format.Format_RGBA8888)

  def to_ps6_pixmap(self):
    """`PySide6.QtGui.QPixmap`"""
    from PySide6.QtGui import QPixmap
    return QPixmap.fromImage(self.to_ps6_image())
  # PyQt5

  def to_qt5_icon(self):
    """`PyQt5.QtGui.QIcon`"""
    from PyQt5.QtGui import QIcon
    return QIcon(self.to_qt5_pixmap())

  def to_qt5_image(self):
    """`PyQt5.QtGui.QImage`"""
    from PyQt5.QtGui import QImage
    img = self.img
    return QImage(img.tobytes(), img.width, img.height, QImage.Format.Format_RGBA8888)

  def to_qt5_pixmap(self):
    """`PyQt5.QtGui.QPixmap`"""
    from PyQt5.QtGui import QPixmap
    return QPixmap.fromImage(self.to_qt5_image())
  # PyQt6

  def to_qt6_icon(self):
    """`PyQt6.QtGui.QIcon`"""
    from PyQt6.QtGui import QIcon
    return QIcon(self.to_qt6_pixmap())

  def to_qt6_image(self):
    """`PyQt6.QtGui.QImage`"""
    from PyQt6.QtGui import QImage
    img = self.img
    return QImage(img.tobytes(), img.width, img.height, QImage.Format.Format_RGBA8888)

  def to_qt6_pixmap(self):
    """`PyQt6.QtGui.QPixmap`"""
    from PyQt6.QtGui import QPixmap
    return QPixmap.fromImage(self.to_qt6_image())



import psutil
import signal
import subprocess
from datetime import timedelta


class Process(psutil.Process):
  def iter_tree(self, include_root=True):
    """Итератор дерева процессов"""
    if include_root:
      yield self
    for p in self.children(True):
      yield p

  def send_signal_tree(self, signal, include_root=True):
    """Отправить сигнал дереву процессов"""
    for p in self.iter_tree(include_root):
      try:
        p.send_signal(signal)
      except psutil.NoSuchProcess:
        pass

  def interrupt_tree(self, include_root=True):
    """Отправить SIGINT дереву процессов"""
    self.send_signal_tree(signal.SIGINT, include_root)

  def kill_tree(self, include_root=True):
    """Убить дерево процессов"""
    for p in self.iter_tree(include_root):
      try:
        p.kill()
      except psutil.NoSuchProcess:
        pass

  def terminate_tree(self, include_root=True):
    """Завершить дерево процессов"""
    for p in self.iter_tree(include_root):
      try:
        p.terminate()
      except psutil.NoSuchProcess:
        pass

  def start_same(self, func=subprocess.Popen, **kw):
    """Запустить такой же процесс"""
    kw.setdefault("args", self.cmdline())
    kw.setdefault("cwd", self.cwd())
    kw.setdefault("env", self.environ())
    if psutil.POSIX:
      kw.setdefault("user", self.username())
    return func(**kw)

  def restart(self, stop_signal=signal.SIGTERM, wait_for_exit: bool | float = True, **kw):
    """Остановить процесс и запустить такой же"""
    self.send_signal(stop_signal)
    if wait_for_exit:
      if isinstance(wait_for_exit, timedelta):
        wait_for_exit = wait_for_exit.total_seconds()
      if isinstance(wait_for_exit, (float, int)):
        # Если число, указать время ожидания
        self.wait(wait_for_exit)
      else:
        # Если True, то ждать без ограничения
        self.wait()
    return self.start_same(**kw)



import pathlib
import shutil
from MainShortcuts2 import ms
from os import fspath
from stat import S_ISDIR, S_ISREG


class Path(pathlib.Path):
  @property
  def ms_path(self):
    """Объект `ms.path.Path`"""
    return ms.path.Path(self)

  def size(self, recursive=False, **walk_kw) -> int:
    """Вычислить размер файла/папки"""
    st = self.stat()
    if S_ISREG(st.st_mode):
      return st.st_size
    if recursive:
      if S_ISDIR(st.st_mode):
        result = 0
        for root, _, filenames in self.walk(**walk_kw):
          for filename in filenames:
            result += (root / filename).size()
        return result
    raise IsADirectoryError(self)

  def walk2(self, **kw):
    """Итерация `(root: Path, dirs: list[Path], files: list[Path])`"""
    for root, dirnames, filenames in self.walk(**kw):
      dirs = [root / i for i in dirnames]
      files = [root / i for i in filenames]
      yield root, dirs, files

  def move_to(self, target):
    """Переместить файл/папку в указанное место"""
    shutil.move(fspath(self), fspath(target))
    return Path(target)

  def copy_to(self, target, follow_symlinks=True, recursive=False, **copytree_kw):
    """Копировать файл/папку в указанное место"""
    if recursive:
      if self.is_dir():
        shutil.copytree(fspath(self), fspath(target), **copytree_kw)
        return Path(target)
    shutil.copy2(fspath(self), fspath(target), follow_symlinks=follow_symlinks)
    return Path(target)

  def remove(self, missing_ok=True, recursive=False, **rmtree_kw):
    """Удалить файл/папку"""
    if recursive:
      if self.is_dir():
        if not self.is_symlink():
          return shutil.rmtree(fspath(self), **rmtree_kw)
    self.unlink(missing_ok=missing_ok)

  def _get_write_path(self):
    """tmp,real"""
    real = self.resolve()
    tmp = real
    while tmp.exists():
      tmp = real.with_name(self.name + ".tmp" + ms.utils.randstr(4))
    return tmp, real

  def read_text(self, **kw) -> str:
    """Прочитать весь текст из файла"""
    kw.setdefault("encoding", "utf-8")
    with self.open("r", **kw) as f:
      return f.read()

  def write_text(self, data: str, **kw) -> int:
    kw.setdefault("encoding", "utf-8")
    """Записать текст в файл"""
    tmp, real = self._get_write_path()
    with tmp.open("w", **kw) as f:
      result = f.write(data)
    if not tmp.samefile(real):
      tmp.replace(real)
    return result

  def write_bytes(self, data, **kw) -> int:
    tmp, real = self._get_write_path()
    with tmp.open("wb", **kw) as f:
      result = f.write(data)
    if not tmp.samefile(real):
      tmp.replace(real)
    return result

  def read_json(self, **kw):
    """Прочитать файл как JSON"""
    return ms.json.read(self, **kw)

  def write_json(self, data, **kw) -> int:
    """Записать JSON в файл"""
    return ms.json.write(self, data, **kw)

  def any_mkdir(self, **kw):
    """Создать папку игнорируя ошибки"""
    self.mkdir(parents=True, exist_ok=True, **kw)
    return self

  def clear_dir(self, mkdir=False, **kw):
    """Очистить папку"""
    if mkdir:
      self.any_mkdir()
    kw.setdefault("recursive", True)
    for i in self.iterdir():
      i.remove(**kw)
    return self

  def read_lines_iter(self, remove_ends=False, **kw):
    """Прочитать строки из файла (итератор)"""
    with self.open("r", **kw) as f:
      while True:
        line = f.readline()
        if line:
          if remove_ends and line.endswith("\n"):
            line = line[:-1]
          yield line
        else:
          break

  def read_lines(self, remove_ends=False, **kw) -> list[str]:
    """Прочитать строки из файла"""
    return list(self.read_lines_iter(remove_ends, **kw))



import sqlite3
# Внимание: возможны SQL-инъекции через названия таблиц и столбцов
# при использовании упрощённых методов (make_*_sql)
# Не указывайте названия из непроверенных источников (например, пользовательский ввод)
TYPE_BY_NAME = {
    "BLOB": bytes,
    "INTEGER": int,
    "NULL": type(None),
    "REAL": float,
    "TEXT": str,
}
NAME_BY_TYPE = {v: k for k, v in TYPE_BY_NAME.items()}


def make_where_string(where: dict, sep=" AND ", prep_null=True):
  params = []
  parts = []
  for k, v in where.items():
    if prep_null and v is None:
      parts.append(k + " IS NULL")
    else:
      params.append(v)
      parts.append(k + "=?")
  return sep.join(parts), params


def make_delete_sql(table: str, where: dict):
  text, params = make_where_string(where)
  return f"DELETE FROM {table} WHERE {text};", params


def make_insert_sql(table: str, values: dict, modifier=""):
  if not values:
    raise ValueError("Cannot insert empty dictionary")
  keys = ",".join(values)
  inserts = ",".join("?" * len(values))
  return f"INSERT {modifier}INTO {table} ({keys}) VALUES ({inserts});", list(values.values())


def make_select_sql(table: str, columns: list[str], where: dict):
  text, params = make_where_string(where)
  return f"SELECT {','.join(columns)} FROM {table} WHERE {text};", params


def make_update_sql(table: str, values: dict, where: dict):
  if not values:
    raise ValueError("No values to update")
  text1, params1 = make_where_string(values, ",", False)
  text2, params2 = make_where_string(where)
  return f"UPDATE {table} SET {text1} WHERE {text2};", params1 + params2


class Cursor(sqlite3.Cursor):
  def __enter__(self):
    return self

  def __exit__(self, *a):
    self.close()

  def execute2(self, sql: str, *params):
    """Удобно если кол-во параметров не меняется"""
    return self.execute(sql, params)
  # Быстрые операции

  def delete(self, table: str, where: dict):
    """Удалить строки из таблицы"""
    return self.execute(*make_delete_sql(table, where))

  def insert_or_ignore(self, table: str, values: dict):
    """Вставить или игнорировать"""
    return self.execute(*make_insert_sql(table, values, "OR IGNORE "))

  def insert_or_replace(self, table: str, values: dict):
    """Вставить или заменить"""
    return self.execute(*make_insert_sql(table, values, "OR REPLACE "))

  def insert(self, table: str, values: dict):
    """Вставить строку в таблицу"""
    return self.execute(*make_insert_sql(table, values))

  def select_fetchall(self, table: str, columns: list[str], where: dict):
    return self.select(table, columns, where).fetchall()

  def select_fetchone(self, table: str, columns: list[str], where: dict):
    return self.select(table, columns, where).fetchone()

  def select(self, table: str, columns: list[str], where: dict):
    """Выбрать строки из таблицы. Не выполняет `fetch`!"""
    return self.execute(*make_select_sql(table, columns, where))

  def update(self, table: str, values: dict, where: dict):
    """Изменить строки в таблице"""
    if not values:
      return self
    return self.execute(*make_update_sql(table, values, where))


class Connection(sqlite3.Connection):
  @classmethod
  def connect(cls, database, **kw):
    """Создать/открыть файл БД"""
    kw.setdefault("check_same_thread", False)
    return cls(database, **kw)

  @classmethod
  def create_in_memory(cls, **kw):
    """Создать БД без файла"""
    return cls.connect(":memory:", **kw)

  def cursor(self):
    return Cursor(self)

  def vacuum(self):
    """Уменьшить размер БД если возможно. Также делает коммит"""
    self.commit()
    with self.cursor() as cur:
      cur.execute("VACUUM;")
  # Методы от курсора

  def delete(self, table: str, where: dict):
    with self.cursor() as cur:
      cur.delete(table, where)

  def execute2(self, sql: str, *params):
    return self.execute(sql, params)

  def insert_or_ignore(self, table: str, values: dict):
    with self.cursor() as cur:
      cur.insert_or_ignore(table, values)

  def insert_or_replace(self, table: str, values: dict):
    with self.cursor() as cur:
      cur.insert_or_replace(table, values)

  def insert(self, table: str, values: dict):
    with self.cursor() as cur:
      cur.insert(table, values)

  def select_fetchall(self, table: str, columns: list[str], where: dict):
    with self.cursor() as cur:
      return cur.select_fetchall(table, columns, where)

  def select_fetchone(self, table: str, columns: list[str], where: dict):
    with self.cursor() as cur:
      return cur.select_fetchone(table, columns, where)

  def update(self, table: str, values: dict, where: dict):
    with self.cursor() as cur:
      cur.update(table, values, where)


connect = Connection.connect
create_in_memory = Connection.create_in_memory



import os
import requests
from .base import BaseClient
from io import BytesIO, IOBase
from MainShortcuts2 import ms
from typing import IO
UPLOAD_HEADERS = {"Content-Type": "application/octet-stream"}


class _ReadBinaryIO(IOBase):
  def __init__(self, resp: requests.Response):
    IOBase.__init__(self)
    resp.raw.decode_content = True
    self.mode: str = "rb"
    self.resp = resp

  def __exit__(self, *a):
    self.close()

  def close(self):
    self.resp.close()

  def read(self, size: int = None) -> bytes:
    return self.resp.raw.read(size)

  def readable(self) -> bool:
    return True

  def seekable(self) -> bool:
    return False

  def writable(self) -> bool:
    return False


class WebDAVClient(BaseClient):
  """Простой WebDAV клиент"""

  def __init__(self, root: str, **kw):
    self._init(**kw)
    self._cwd: str = "/"
    self._url = root.rstrip("/") + "{method}"

  def _auth(self):
    pass  # Без авторизации

  def _fspath(self, path) -> str:
    if isinstance(path, bytes):
      path = path.decode("utf-8")
    return os.fspath(path)

  def resolve_path(self, path: str) -> str:
    if not isinstance(path, str):
      path = self._fspath(path)
    if not path.startswith("/"):
      path = self.cwd + "/" + path
    return os.path.normpath(path).replace("\\", "/")

  @property
  def cwd(self) -> str:
    """Рабочая папка на сервере"""
    return self._cwd

  @cwd.setter
  def cwd(self, value: str):
    self._cwd = self.resolve_path(value).rstrip("/")

  def request(self, method: str, path: str, **kw) -> requests.Response:
    self._auth()
    path = self.resolve_path(path)
    if not kw.get("stream"):
      with self._request(method, path, **kw) as resp:
        return resp
    return self._request(method, path, **kw)

  def open(self, path: str, mode: str = "rb") -> _ReadBinaryIO:
    """Открыть файл сервера (read-only)"""
    if not mode in ["rb", "br"]:
      raise ValueError("Only the 'rb' mode is supported")
    return _ReadBinaryIO(self.request("GET", path, stream=True))

  def download(self, path: str, filepath: str, **kw) -> int:
    """Скачать файл с сервера"""
    with self.request("GET", path, stream=True) as resp:
      return ms.utils.download_file(resp, filepath)

  def upload_io(self, io: IO[bytes], path: str, **kw) -> requests.Response:
    if isinstance(io, bytes):
      with BytesIO(io) as f:
        return self.upload_io(f, path, **kw)
    kw.setdefault("headers", {})
    for k, v in UPLOAD_HEADERS.items():
      kw["headers"].setdefault(k, v)
    if not kw["headers"].get("Content-Length"):
      if io.seekable():
        pos = io.tell()
        io.seek(0, 2)
        kw["headers"]["Content-Length"] = str(io.tell())
        io.seek(pos)
    return self.request("PUT", path, data=io, **kw)

  def upload(self, filepath: str, path: str, **kw):
    """Загрузить локальный файл на сервер"""
    with open(filepath, "rb") as f:
      return self.upload_io(f, path, **kw)

  def mkdir(self, path: str):
    """Создать пустую папку на сервере"""
    if not path.endswith("/"):
      path += "/"
    return self.request("MKCOL", path)

  def delete(self, path: str):
    """Удалить файл/папку на сервере"""
    return self.request("DELETE", path)

  def move(self, src: str, dst: str, overwrite=False, **kw):
    """Переместить файл/папку на сервере"""
    kw.setdefault("headers", {})
    kw["headers"]["Destination"] = self.resolve_path(dst)
    kw["headers"]["Overwrite"] = "T" if overwrite else "F"
    return self.request("MOVE", src, **kw)

  def copy(self, src: str, dst: str, overwrite=False, **kw):
    """Скопировать файл/папку на сервере"""
    kw.setdefault("headers", {})
    kw["headers"]["Destination"] = self.resolve_path(dst)
    kw["headers"]["Overwrite"] = "T" if overwrite else "F"
    return self.request("COPY", src, **kw)

  def rename(self, src: str, name: str, **kw):
    """Переименовать файл/папку на сервере"""
    parent = os.path.dirname(self.resolve_path(src))
    return self.move(src, parent + "/" + name, **kw)



from MainShortcuts2.ex.pathlib_ex import Path
from MainShortcuts2.gui_scripts.utils import *
from MainShortcuts2.ms2hash import *
argp = ArgumentParser()
argp.add_argument("files", nargs="*", help="пути к файлам")
argp.add_argument("-f", "--force", action="store_true", help="перезаписывать существующие хеши")
argp.add_argument("-t", "--type", choices=HASH_TYPES, default="sha512", help="тип контрольной суммы")


def prep_files(paths: list[str]) -> set[Path]:
  result = set()
  for str_path in paths:
    obj_path = Path(str_path)
    suffix = HASH_SUFFIX.lower()
    while obj_path.suffix.lower() == suffix:
      obj_path = obj_path.with_suffix("")
    if obj_path.is_file():
      result.add(obj_path)
  return result


class MainWindow(QMainWindow):
  def __init__(self, files: set[Path], force=False, hash_type="sha512"):
    super().__init__()
    self.files = files
    self.force = force
    self.hash_type = hash_type
    # Настройка окна


raise Exception("Модуль в разработке")


@main_func(__name__)
def main(args=None, *, qapp: QApplication):
  if args is None:
    args = argp.parse_args()
  files = prep_files(args.files)
  if not files:
    mw = QMessageBox()
    mw.setIcon(QMessageBox.Icon.Critical)
    mw.setText("Нет доступных файлов для обработки")
    mw.setWindowTitle("Ошибка")
    return mw.exec()



"""Работа с объектами файловой системы и путями к ним"""
import io
import os
import pathlib
import shutil
import stat
import sys
from .core import ms
from datetime import datetime
from functools import cached_property
from typing import *
EXTSEP = "."
FORBIDDEN_SYMBOLS = [":", "!", "?", "@", "*", "\"", "\n", "%", "+", "<", ">", "|"]
PATH_TYPES = Union[io.FileIO, pathlib.Path, str]
PATHSEP = "/"


def _path2str(path) -> str:
  if isinstance(path, str):
    return path
  if isinstance(path, pathlib.Path):
    return path.as_posix()
  if isinstance(path, io.FileIO):
    return path.name
  if isinstance(path, bytes):
    return path.decode()
  return os.fspath(path)


def path2str(path: PATH_TYPES, to_abs: bool = False, replace_forbidden_to: str = None) -> str:
  """Преобразовать объект Python в строковый путь"""
  result = _path2str(path)
  if not replace_forbidden_to is None:
    for i in FORBIDDEN_SYMBOLS:
      if i in replace_forbidden_to:
        raise ValueError("It is impossible to replace the forbidden symbol with another forbidden symbol")
      result = result.replace(i, replace_forbidden_to)
  if to_abs:
    result = os.path.abspath(result)
  else:
    result = os.path.normpath(result)
  result = result.replace("\\", PATHSEP)
  if len(result) > 3 and result[-1] == "/":
    result = result[:-1]
  return result


def cwd(set_to: PATH_TYPES = None, only_set=False) -> str:
  """Получить путь к рабочей папке"""
  if not set_to is None:
    os.chdir(path2str(set_to))
  if not only_set:
    return os.getcwd().replace("\\", PATHSEP)


class Path(os.PathLike):
  """Информация и действия с объектом файловой системы"""
  TYPE_DIR = "dir"
  TYPE_FILE = "file"
  TYPE_NOT_EXISTS = "not_exists"
  TYPE_UNKNOWN = "unknown"

  def __init__(self, path: PATH_TYPES, use_cache: bool = True):
    self._path = path2str(path, to_abs=True)
    self.use_cache = use_cache
    self.reload(full=True)

  def __fspath__(self) -> str:
    return self.path

  def __repr__(self):
    cls = type(self)
    if self.use_cache:
      return cls.__module__ + "." + cls.__name__ + "(" + repr(self.path) + ")"
    return cls.__module__ + "." + cls.__name__ + "(" + repr(self.path) + ", use_cache = False)"

  def __str__(self):
    return self.path

  def __lt__(self, other):
    return self.path < path2str(other, to_abs=True)

  def __add__(self, other) -> "Path":
    if isinstance(other, str):
      return Path(self.path + other, use_cache=self.use_cache)
    return NotImplemented

  def __truediv__(self, other) -> "Path":
    if isinstance(other, str):
      return Path(os.path.join(self.path, other), use_cache=self.use_cache)
    return NotImplemented

  def reload(self, full: bool = False):
    """Удаление кешированной информации"""
    if full:
      self._base_name = None
      self._ext = None
      self._full_name = None
      self._parent_dir = None
      self._split = None
    self._created_at = None
    self._exists = None
    self._is_dir = None
    self._is_file = None
    self._is_link = None
    self._modified_at = None
    self._realpath = None
    self._size = None
    self._type = None
    self._used_at = None

  def _stat(self):
    st = os.stat(self.path)
    self._created_at = st.st_ctime
    self._exists = True
    self._is_dir = stat.S_ISDIR(st.st_mode)
    self._is_file = stat.S_ISREG(st.st_mode)
    self._modified_at = st.st_mtime
    self._size = st.st_size
    self._used_at = st.st_atime
    if self._is_dir:
      self._type = self.TYPE_DIR
    elif self._is_file:
      self._type = self.TYPE_FILE
    return st

  @property
  def path(self) -> str:
    """Абсолютный путь к объекту"""
    return self._path

  @path.setter
  def path(self, v):
    """Абсолютный путь к объекту"""
    self._path = path2str(v, to_abs=True)
    self.reload(full=True)

  @property
  def base_name(self) -> str:
    """Имя объекта без расширения (`example.txt` -> `example`)"""
    if self._base_name is None:
      self._base_name, self._ext = os.path.splitext(self.full_name)
    return self._base_name

  @property
  def created_at(self) -> float:
    """timestamp создания объекта"""
    if self._created_at is None or (not self.use_cache):
      self._stat()
    return self._created_at

  @property
  def exists(self) -> bool:
    """Существует ли объект"""
    if self._exists is None or (not self.use_cache):
      try:
        self._stat()
      except (OSError, ValueError):
        self._exists = False
    return self._exists

  @property
  def ext(self) -> str:
    """Расширение объекта включая точку (`example.txt` -> `.txt`)"""
    if self._ext is None:
      self._base_name, self._ext = os.path.splitext(self.full_name)
    return self._ext

  @property
  def full_name(self) -> str:
    """Полное имя объекта включая расширение (`/home/example.txt` -> `example.txt`)"""
    if self._full_name is None:
      self._parent_dir, self._full_name = os.path.split(self.path)
    return self._full_name

  @property
  def is_dir(self) -> bool:
    """Является ли объект папкой"""
    if self._is_dir is None or (not self.use_cache):
      try:
        self._stat()
      except (OSError, ValueError):
        self._exists = False
        self._is_dir = False
        return False
    return self._is_dir

  @property
  def is_file(self) -> bool:
    """Является ли объект файлом"""
    if self._is_file is None or (not self.use_cache):
      try:
        self._stat()
      except (OSError, ValueError):
        self._exists = False
        self._is_file = False
        return False
    return self._is_file

  @property
  def is_link(self) -> bool:
    """Является ли объект ссылкой на другой объект"""
    if self._is_link is None or (not self.use_cache):
      self._is_link = os.path.islink(self.path)
    return self._is_link

  @property
  def modified_at(self) -> float:
    """timestamp изменения объекта"""
    if self._modified_at is None or (not self.use_cache):
      self._stat()
    return self._modified_at

  @property
  def parent_dir(self) -> str:
    """Путь к родительской папке (`/home/example.txt` -> `/home`)"""
    if self._parent_dir is None:
      self._parent_dir, self._full_name = os.path.split(self.path)
    return self._parent_dir

  @property
  def realpath(self) -> str:
    """Настоящий путь к объекту, если это ссылка (может быть неправильным)"""
    if self._realpath is None or (not self.use_cache):
      self._realpath = os.path.realpath(self.path)
    return self._realpath

  @property
  def size(self) -> int:
    """Размер объекта в байтах"""
    if self._size is None or (not self.use_cache):
      self._stat()
    return self._size

  @property
  def split(self) -> tuple:
    """Разбитый путь к объекту"""
    if self._split is None:
      self._split = self.path.split(PATHSEP)
    return self._split

  @property
  def type(self) -> str:
    """Тип объекта (`dir` | `file`)"""
    if self._type is None or (not self.use_cache):
      if self.is_dir:
        self._type = self.TYPE_DIR
      elif self.is_file:
        self._type = self.TYPE_FILE
      elif not self.exists:
        self._type = self.TYPE_NOT_EXISTS
      else:
        self._type = self.TYPE_UNKNOWN
    return self._type

  @property
  def used_at(self) -> float:
    """timestamp последнего использования объекта"""
    if self._used_at is None or (not self.use_cache):
      self._stat()
    return self._used_at

  @cached_property
  def pathlib_path(self):
    from MainShortcuts2.ex.pathlib_ex import Path
    return Path(str(self))

  def copy(self, dest: PATH_TYPES, follow: bool = False, **kw) -> str:
    """Копирование объекта"""
    r = copy(self.path, dest, **kw)
    if follow:
      self.path = r
    return r

  def delete(self, **kw):
    """Безвозвратное удаление объекта"""
    delete(self.path, **kw)
    self.reload()

  def in_dir(self, dir: PATH_TYPES, **kw) -> bool:
    """Находится ли объект в папке"""
    return in_dir(self.path, dir, **kw)

  def link(self, dest: PATH_TYPES, follow: bool = False, **kw) -> str:
    """Создать символическую ссылку на объект"""
    r = link(self.path, dest, **kw)
    if follow:
      self.path = r
    return r

  def move(self, dest: PATH_TYPES, follow: bool = True, **kw) -> str:
    """Переместить объект"""
    r = move(self.path, dest, **kw)
    if follow:
      self.path = r
    return r

  def rename(self, name: str, follow: bool = True, **kw) -> str:
    """Переименовать объект"""
    r = rename(self.path, name, **kw)
    if follow:
      self.path = r
    return r

  def list_dir(self, **kw):
    """Список содержимого папки"""
    return ms.dir.list(self, **kw)

  def to_dict(self, hashes: list[str] = None, empty_values=True) -> dict:
    """Создать словарь с данными объекта (совместим с JSON)"""
    result = {}
    result["path"] = self.path
    if self.exists:
      result["created_at"] = self.created_at
      result["exists"] = True
      result["is_link"] = self.is_link
      result["modified_at"] = self.modified_at
      result["realpath"] = self.realpath
      result["type"] = self.type
      result["used_at"] = self.used_at
      if self.is_file:
        result["size"] = self.size
      if hashes:
        result["hashes"] = self.multi_hash_hex(hashes)
    else:
      result["exists"] = False
    if empty_values:
      for i in ("created_at", "hashes", "is_link", "modified_at", "realpath", "type", "used_at"):
        result.setdefault(i, None)
    return result

  def list_dir_iter(self, **kw):
    """Итератор по списку содержимого папки"""
    return ms.dir.list_iter(self, **kw)

  def open_file(self, mode: str = "r", **kw):
    """Открыть файл на чтение/запись"""
    return open(self.path, mode, **kw)

  def hash(self, algorithm: str, chunk_size=ms.file.CHUNK_SIZE) -> bytes:
    """Хеш файла (`bytes`)"""
    import hashlib
    hash = hashlib.new(algorithm)
    with self.open_file("rb") as f:
      rf = f.read  # read func
      uf = hash.update  # update func
      while True:
        b = rf(chunk_size)
        if not b:
          break
        uf(b)
    return hash.digest()

  def hash_hex(self, algorithm: str, **kw) -> str:
    """Хеш файла (`bytes.hex()`)"""
    return self.hash(algorithm, **kw).hex()

  def hash_b85(self, algorithm: str, **kw) -> str:
    """Хеш файла (`base85`)"""
    from base64 import b85encode
    return b85encode(self.hash(algorithm, **kw)).decode()

  def multi_hash(self, algorithms: list[str], chunk_size=ms.file.CHUNK_SIZE) -> dict[str, bytes]:
    """Хеш файла (`bytes`) для нескольких алгоритмов"""
    if not algorithms:
      return {}
    import hashlib
    hashes = {i: hashlib.new(i) for i in set(algorithms)}
    ufs = [i.update for i in hashes.values()]
    with self.open_file("rb") as f:
      rf = f.read  # read func
      while True:
        b = rf(chunk_size)
        if not b:
          break
        for uf in ufs:
          uf(b)
    return {k: v.digest() for k, v in hashes.items()}

  def multi_hash_hex(self, algorithms: list[str], **kw) -> dict[str, str]:
    """Хеш файла в HEX строке для нескольких алгоритмов"""
    return {k: v.hex() for k, v in self.multi_hash(algorithms, **kw).items()}

  def multi_hash_b85(self, algorithms: list[str], **kw) -> dict[str, str]:
    """Хеш файла в base85 строке для нескольких алгоритмов"""
    from base64 import b85encode
    result = {}
    for k, v in self.multi_hash(algorithms, **kw).items():
      result[k] = b85encode(v).decode()
    return result

  def to_pathlib(self):
    """Создать объект `pathlib.Path` из этого пути"""
    return pathlib.Path(self.path)

  def same_file(self, other):
    """Является ли другой объект тем же файлом"""
    if isinstance(other, os.stat_result):
      return os.path.samestat(self._stat(), other)
    return os.path.samefile(self.path, other)


class Stat:
  """Состояние объекта на диске"""

  def __init__(self, data: os.stat_result):
    self._atime_dt = None
    self._ctime_dt = None
    self._mtime_dt = None
    self.data = data
    self.atime: float = self.data.st_atime
    self.ctime: float = self.data.st_birthtime if sys.version_info >= (3, 12) and sys.platform == "win32" else self.data.st_ctime
    self.dev: int = self.data.st_dev
    self.group_id: int = self.data.st_gid
    self.inode: int = self.data.st_ino
    self.is_dir: bool = stat.S_ISDIR(self.data.st_mode)
    self.is_file: bool = stat.S_ISREG(self.data.st_mode)
    self.mtime: float = self.data.st_mtime
    self.size: int = self.data.st_size
    self.user_id: int = self.data.st_uid

  def __eq__(self, other):
    if isinstance(other, Stat):
      return os.path.samestat(self.data, other.data)
    if isinstance(other, os.stat_result):
      return os.path.samestat(self.data, other)
    return NotImplemented

  @property
  def atime_dt(self) -> datetime:
    """Дата последнего доступа к файлу"""
    if self._atime_dt is None:
      self._atime_dt = datetime.fromtimestamp(self.atime)
    return self._atime_dt

  @property
  def ctime_dt(self) -> datetime:
    """Дата создания файла"""
    if self._ctime_dt is None:
      self._ctime_dt = datetime.fromtimestamp(self.ctime)
    return self._ctime_dt

  @property
  def mtime_dt(self) -> datetime:
    """Дата изменения файла"""
    if self._mtime_dt is None:
      self._mtime_dt = datetime.fromtimestamp(self.mtime)
    return self._mtime_dt

  @classmethod
  def load(cls, path: str, **kw):
    """Получить из локального файла"""
    return cls(os.stat(path, **kw))

  @classmethod
  def lload(cls, path: str, **kw):
    return cls(os.lstat(path, **kw))


def copy(path: PATH_TYPES, dest: PATH_TYPES, mkdir: bool = False, **kw) -> str:
  """Копировать объект"""
  kw["dst"] = path2str(dest)
  kw["src"] = path2str(path)
  if mkdir:
    ms.dir.create(os.path.dirname(kw["dst"]))
  if os.path.isfile(kw["src"]):
    shutil.copy2(**kw)
    return kw["dst"]
  if os.path.isdir(kw["src"]):
    shutil.copytree(**kw)
    return kw["dst"]


def delete(path: PATH_TYPES, **kw):
  """Удалить объект с содержимым"""
  kw["path"] = path2str(path)
  if os.path.islink(kw["path"]):
    return os.unlink(**kw)
  if os.path.isdir(kw["path"]):
    return shutil.rmtree(**kw)
  if os.path.isfile(kw["path"]):
    return os.remove(**kw)


def exists(path: PATH_TYPES) -> bool:
  """Существует ли объект"""
  return os.path.exists(path2str(path))


def in_dir(path: str, dir: str, recursion_limit: int = 1000) -> bool:
  """Находится ли объект в указанной папке"""
  dir = path2str(dir, True)
  path = path2str(path, True)
  while True:
    pdir = os.path.dirname(path)
    if path == pdir:
      return False
    if dir == pdir:
      return True
    path = pdir
    recursion_limit -= 1
    if recursion_limit < 0:
      raise RecursionError()


def is_dir(path: PATH_TYPES) -> bool:
  """Является ли объект папкой"""
  return os.path.isdir(path2str(path))


def is_file(path: PATH_TYPES) -> bool:
  """Является ли объект файлом"""
  return os.path.isfile(path2str(path))


def is_link(path: PATH_TYPES) -> bool:
  """Является ли объект ссылкой на другой объект"""
  return os.path.islink(path2str(path))


def link(path: PATH_TYPES, dest: PATH_TYPES, force: bool = False, mkdir: bool = False, **kw) -> str:
  """Сделать символическую ссылку на объект"""
  kw["dst"] = path2str(dest, True)
  kw["src"] = path2str(path, True)
  if force:
    mkdir = True
    if exists(kw["dst"]):
      delete(kw["dst"])
  if mkdir:
    ms.dir.create(os.path.dirname(kw["dst"]))
  os.symlink(**kw)
  return kw["dst"]


def move(path: PATH_TYPES, dest: PATH_TYPES, force: bool = False, mkdir: bool = False, **kw) -> str:
  """Переместить объект"""
  kw["dst"] = path2str(dest)
  kw["src"] = path2str(path)
  if force:
    mkdir = True
    if exists(kw["dst"]):
      delete(kw["dst"])
  if mkdir:
    ms.dir.create(os.path.dirname(kw["dst"]))
  shutil.move(**kw)
  return kw["dst"]


def rename(path: PATH_TYPES, name: PATH_TYPES, **kw) -> str:
  """Переименовать объект"""
  kw["path"] = path
  kw["dest"] = os.path.dirname(path2str(path)) + "/" + name
  return move(**kw)


class TempFiles:
  """Удаление временных файлов по окончании операций с `with`"""

  def __init__(self, *files: PATH_TYPES):
    self.files: list[str] = []
    self.add(*files)

  def __enter__(self):
    return self

  def __exit__(self, a, b, c):
    self.remove_files()

  def __contains__(self, k):
    return k in self.files

  def __delitem__(self, k):
    if isinstance(k, str):
      k = self.files.index(k)
    del self.files[k]

  def add(self, *files: PATH_TYPES):
    """Добавить файлы в список временных"""
    for i in files:
      path = path2str(i, to_abs=True)
      if not path in self.files:
        self.files.append(path)

  def remove_files(self, ignore_errors: bool = True):
    """Удалить все временные файлы"""
    for i in self.files:
      try:
        delete(i)
      except Exception:
        if not ignore_errors:
          raise


def _rec_readlink(path: str, hist: list[str] = None) -> str:
  if hist is None:
    hist = [path]
  while os.path.islink(path):
    path = os.path.realpath(path).replace("\\", "/")
    if path in hist:
      raise RecursionError()
    hist.append(path)
  return path


def readlink(path: PATH_TYPES, recursive: bool = False) -> str:
  path = path2str(path, True)
  if recursive:
    return _rec_readlink(path)
  if not os.path.islink(path):
    return path
  return os.path.realpath(path)


cp = copy
ln = link
mv = move
rm = delete
rn = rename



import typing
from aiohttp import BasicAuth, hdrs, web
from aiohttp.typedefs import Handler
from aiohttp.web import FileResponse, Request, Response, StreamResponse
from logging import DEBUG
from MainShortcuts2 import ms
from pathlib import Path


def wrap_handler(app: "Application", handler: Handler):
  async def wrapper(req: Request) -> StreamResponse | typing.Any:
    try:
      resp = await handler(req)
      if isinstance(resp, int):
        resp = Response(status=resp)
      elif isinstance(resp, Path):
        resp = FileResponse(resp)
    except Exception as exc:
      app.logger.exception("Failed to process request", exc_info=exc)
      resp = Response(status=500, reason="Internal server error")
    if app.logger.isEnabledFor(DEBUG):
      remote = req.remote or "unknown"
      if "X-Real-IP" in req.headers:
        remote += " (real IP: %s)" % req.headers["X-Real-IP"]
      status = resp.status if isinstance(resp, StreamResponse) else repr(resp)
      app.logger.debug("Request %s %s from %s, return %s", req.method, req.raw_path, remote, status)
    return resp
  return wrapper


class Application(web.Application):
  def run(self, **kw):
    """Запустить приложение"""
    kw.setdefault("print", self.logger.info)
    web.run_app(self, **kw)

  def run_local(self, port: int, **kw):
    """Запустить локально на указанном порту"""
    self.run(host="127.0.0.1", port=port, **kw)

  def on_request(self, methods: str | list[str], paths: str | list[str], add_slash=True, **kw):
    """Зарегистрировать обработчик запросов"""
    if isinstance(methods, str):
      methods = [methods]
    if isinstance(paths, str):
      paths = [paths]

    def deco(func: Handler):
      wrapper = wrap_handler(self, func)
      for m in methods:
        for p in paths:
          self._on_request(wrapper, m, p, add_slash, **kw)
      return func
    return deco

  def _on_request(self, func: Handler, method: str, path: str, add_slash=True, **kw):
    if path == "/":
      add_slash = False
    if add_slash:
      path = path.rstrip("/")
    self.router.add_route(method, path, func, **kw)
    if add_slash:
      self.router.add_route(method, path + "/", func, **kw)

  def on_any_path_request(self, methods: str | list[str], **kw):
    """Зарегистрировать обработчик запросов на любой путь"""
    return self.on_request(methods, "/{tail:.*}", False, **kw)


def auth_header(header=hdrs.AUTHORIZATION, value=None):
  """Простая авторизация заголовком. Если не указать требуемое значение, будет проверять только наличие заголовка"""
  def deco(func: Handler):
    async def wrapper(req: Request):
      if header in req.headers:
        if value is not None:
          if req.headers[header] != value:
            return Response(status=403)  # Заголовок не совпадает
        return await func(req)  # Заголовок совпадает или значение не указано
      return Response(status=401)  # Заголовок отсутствует
    return wrapper
  return deco


basic_unauthorized_headers = {"WWW-Authenticate": 'Basic realm="Basic Auth"'}


def auth_basic(header=hdrs.AUTHORIZATION, filter: typing.Callable[[BasicAuth], bool] = ms.utils.return_True):
  """Авторизация `Basic`. Если не указать фильтр, только проверяет наличие заголовка"""
  def deco(func: typing.Callable[[Request, BasicAuth], typing.Awaitable[StreamResponse]]):
    async def wrapper(req: Request):
      if header in req.headers:
        try:
          auth = BasicAuth.decode(req.headers[header])
        except Exception:
          return Response(status=400, headers=basic_unauthorized_headers)  # Неправильный заголовок
        if filter(auth):
          return await func(req, auth)  # Одобрено
        return Response(status=403)  # Запрещено
      return Response(status=401, headers=basic_unauthorized_headers)  # Заголовок отсутствует
    return wrapper
  return deco
